# -*- coding: utf-8 -*-
"""Synteny (genome order) operations."""
# standard library imports
import sys
from itertools import combinations

# from os.path import commonprefix as prefix
from pathlib import Path

# third-party imports
import dask.bag as db
import networkx as nx
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar

# module imports
from .common import AMBIGUOUS_CODE
from .common import ANCHOR_HIST_FILE
from .common import ANCHORS_FILE
from .common import CLUSTERS_FILE
from .common import CLUSTERSYN_FILE
from .common import CODE_DICT
from .common import DISAMBIGUATED_CODE
from .common import HOMOLOGY_FILE
from .common import INDIRECT_CODE
from .common import LOCALLY_UNAMBIGUOUS_CODE
from .common import NON_AMBIGUOUS_CODE
from .common import PROTEOMOLOGY_FILE
from .common import PROTEOSYN_FILE
from .common import SPINNER_UPDATE_PERIOD
from .common import SYNTENY_FILE
from .common import SYNTENY_FILETYPE
from .common import UNAMBIGUOUS_CODE
from .common import calculate_adjacency_group
from .common import dotpath_to_path
from .common import hash_array
from .common import log_and_add_to_stats
from .common import logger
from .common import read_tsv_or_parquet
from .common import write_tsv_or_parquet
from .hash import SyntenyBlockHasher
from .mailboxes import DataMailboxes
from .mailboxes import ExternalMerge
from .merger import AmbiguousMerger

# global constants
__ALL__ = ["synteny_anchors"]
CLUSTER_COLS = ["syn.anchor.id", "syn.anchor.count", "syn.anchor.direction"]
JOIN_COLS = [
    "member_ids",
    "syn.anchor.sub_id",
    "syn.anchor.id",
    "syn.anchor.count",
    "syn.code",
    "frag.idx",
]
ANCHOR_COLS = [
    "path",
    "syn.code",
    "syn.anchor.count",
    "hom.cluster",
    "frag.id",
    "frag.pos",
    "hom.cl_size",
    "frag.direction",
    "frag.idx",
    "frag.is_chr",
    "frag.is_plas",
    "frag.is_scaf",
    "frag.prot_count",
    "frag.start",
    "prot.len",
    "prot.m_start",
    "prot.n_ambig",
    "prot.no_stop",
]
MAILBOX_SUBDIR = "mailboxes"

# CLI function
def synteny_anchors(
    k,
    peatmer,
    setname,
    click_loguru=None,
    write_ambiguous=True,
    thorny=True,
    disambig_adj_only=True,
):
    """Calculate synteny anchors."""
    #
    # Marshal input arguments
    #
    if k < 2:
        logger.error("k must be at least 2.")
        sys.exit(1)
    options = click_loguru.get_global_options()
    user_options = click_loguru.get_user_global_options()
    set_path = Path(setname)
    file_stats_path = set_path / PROTEOMOLOGY_FILE
    proteomes = read_tsv_or_parquet(file_stats_path)
    n_proteomes = len(proteomes)
    clusters = read_tsv_or_parquet(set_path / CLUSTERS_FILE)
    n_clusters = len(clusters)
    hasher = SyntenyBlockHasher(
        k=k,
        peatmer=peatmer,
        thorny=thorny,
        disambig_adj_only=disambig_adj_only,
    )
    logger.info(
        f"Calculating {hasher.hash_name(no_prefix=True)} synteny anchors"
        + f" for {n_proteomes} proteomes"
    )
    # durable argument list for passes
    arg_list = [
        (
            idx,
            row["path"],
        )
        for idx, row in proteomes.iterrows()
    ]
    runner = PassRunner(
        {
            "n_proteomes": n_proteomes,
            "set_path": set_path,
            "hasher": hasher,
            "quiet": options.quiet,
            "parallel": user_options["parallel"],
            "bag": db.from_sequence(arg_list),
            "merge_args": arg_list,
            "click_loguru": click_loguru,
        }
    )
    #
    # Make first three passes:
    #  1. hash and find unambiguous anchors
    #  2. disambiguate ambiguous anchors adjacent to unambiguous ones
    #  3. find non-ambiguous hashes uncovered by 2)
    #
    for pass_code in [
        UNAMBIGUOUS_CODE,
        DISAMBIGUATED_CODE,
        NON_AMBIGUOUS_CODE,
    ]:
        proteomes = runner.make_pass(pass_code, proteomes)
    runner.add_ambig_to_total_assigned()
    n_anchors = runner.get_total_assigned()
    #
    # Fourth pass -- merge, write anchor and homology info
    #
    join_mb = DataMailboxes(
        n_boxes=n_proteomes,
        mb_dir_path=(set_path / MAILBOX_SUBDIR / "join"),
        file_extension="tsv",
    )
    join_mb.write_tsv_headers(JOIN_COLS)
    cluster_mb = DataMailboxes(
        n_boxes=n_clusters,
        mb_dir_path=(set_path / MAILBOX_SUBDIR / "clusters"),
        file_extension="tsv",
    )
    cluster_mb.write_tsv_headers(CLUSTER_COLS)
    anchor_mb = DataMailboxes(
        n_boxes=n_anchors,
        mb_dir_path=(set_path / MAILBOX_SUBDIR / "anchors"),
        file_extension="tsv",
    )
    anchor_mb.write_tsv_headers(ANCHOR_COLS)
    proteomes = runner.make_pass(
        INDIRECT_CODE,
        proteomes,
        extra_kwargs={
            "join_mb": join_mb,
            "cluster_mb": cluster_mb,
            "anchor_mb": anchor_mb,
            "n_proteomes": n_proteomes,
            "write_ambiguous": write_ambiguous,
        },
    )
    write_tsv_or_parquet(
        proteomes, set_path / PROTEOSYN_FILE, remove_tmp=False
    )
    adjacency_stats = anchors_to_adjacency(
        set_path, n_proteomes, join_mb.open_then_delete
    )
    logger.info(f"adjacency_stats: {adjacency_stats}")
    return
    #
    # Write anchors
    #
    click_loguru.elapsed_time("Anchor writing")
    arg_list = [(i,) for i in range(n_anchors)]
    anchor_path = set_path / "synteny"
    anchor_path.mkdir(exist_ok=True)
    logger.info(f"Writing {n_anchors} synteny anchors to {anchor_path}:")
    if not options.quiet:
        ProgressBar(dt=SPINNER_UPDATE_PERIOD).register()
    if parallel:
        bag = db.from_sequence(arg_list)
        anchor_stats = bag.map(
            write_anchor,
            mailbox_reader=anchor_mb.open_then_delete,
            synteny_parent=anchor_path,
        ).compute()
    else:
        anchor_stats = []
        for args in arg_list:
            anchor_stats.append(
                write_anchor(
                    args,
                    mailbox_reader=anchor_mb.open_then_delete,
                    synteny_parent=anchor_path,
                )
            )
    anchor_mb.delete()
    anchor_stat_list = []
    for results in anchor_stats:
        if results is not None:
            anchor_stat_list += results
    anchor_frame = pd.DataFrame.from_dict(results)
    write_tsv_or_parquet(
        anchor_frame,
        set_path / ANCHORS_FILE,
        sort_cols=False,
    )
    #
    # Merge synteny into clusters
    #
    arg_list = [(i,) for i in range(n_clusters)]
    click_loguru.elapsed_time("Synteny joining")
    homology_path = set_path / "homology"
    logger.info(
        f"Joining synteny info to {n_clusters} clusters in {homology_path}:"
    )
    if not options.quiet:
        ProgressBar(dt=SPINNER_UPDATE_PERIOD).register()
    if parallel:
        bag = db.from_sequence(arg_list)
        cluster_stats = bag.map(
            join_synteny_to_clusters,
            mailbox_reader=cluster_mb.open_then_delete,
            cluster_parent=homology_path,
        ).compute()
    else:
        cluster_stats = []
        for args in arg_list:
            cluster_stats.append(
                join_synteny_to_clusters(
                    args,
                    mailbox_reader=cluster_mb.open_then_delete,
                    cluster_parent=homology_path,
                )
            )
    cluster_mb.delete()
    cluster_frame = pd.DataFrame.from_dict(cluster_stats)
    cluster_frame.set_index(["clust_id"], inplace=True)
    cluster_frame.sort_index(inplace=True)
    clusters = _concat_without_overlap(clusters, cluster_frame)
    write_tsv_or_parquet(
        clusters, set_path / CLUSTERSYN_FILE, float_format="%5.2f"
    )
    mean_gene_synteny = (
        clusters["in_synteny"].sum() * 100.0 / clusters["size"].sum()
    )
    mean_clust_synteny = clusters["synteny_pct"].mean()
    logger.info(
        f"Mean anchor coverage: {mean_gene_synteny: .1f}% (on proteins)"
    )
    logger.info(
        f"Mean cluster anchor coverage: {mean_clust_synteny:.1f}% (on clusters)"
    )
    click_loguru.elapsed_time(None)


class PassRunner:
    """Run a pass over all proteomes."""

    def __init__(self, std_kwargs):
        """Save initial pass info"""
        self.std_kwargs = std_kwargs
        self.last_code = None
        self.pass_name = "Hashing"
        self.merger_kw_dict = {
            UNAMBIGUOUS_CODE: {
                "count_key": "syn.anchor.count",
                "ordinal_key": "syn.anchor.id",
                "ambig_ordinal_key": "tmp.ambig.id",
            },
            DISAMBIGUATED_CODE: {
                "count_key": "tmp.disambig.anchor.count",
                "ordinal_key": "tmp.disambig.anchor.id",
                "alt_hash": True,
            },
            NON_AMBIGUOUS_CODE: {
                "count_key": "tmp.nonambig.anchor.count",
                "ordinal_key": "tmp.nonambig.anchor.id",
                "ambig_count_key": "tmp.ambig.anchor.count",
                "ambig_ordinal_key": "tmp.ambig.anchor.id",
            },
        }
        self.merge_function_dict = {
            UNAMBIGUOUS_CODE: calculate_synteny_hashes,
            DISAMBIGUATED_CODE: merge_unambig_hashes,
            NON_AMBIGUOUS_CODE: merge_disambig_hashes,
            INDIRECT_CODE: merge_nonambig_hashes,
        }
        self.n_assigned_list = []
        self.ambig = None
        self.unambig = None
        self.log_ambig = False

    def make_pass(
        self,
        code,
        proteomes,
        extra_kwargs=None,
    ):
        """Make a calculate-merge pass over each proteome."""
        self.std_kwargs["click_loguru"].elapsed_time(self.pass_name)
        if self.unambig is not None:
            if self.log_ambig:
                ambig_msg = f" and {len(self.ambig)} ambiguous"
            else:
                ambig_msg = ""
            logger.info(
                f"Merging {len(self.unambig)} {CODE_DICT[self.last_code]}"
                + f"({self.last_code}){ambig_msg} synteny anchors into proteomes"
            )
        if extra_kwargs is None:
            extra_kwargs = {}
        kwargs = {
            "unambig": self.unambig,
            "ambig": self.ambig,
            "hasher": self.std_kwargs["hasher"],
        }
        if code in self.merger_kw_dict:
            mailboxes = DataMailboxes(
                n_boxes=self.std_kwargs["n_proteomes"],
                mb_dir_path=(
                    self.std_kwargs["set_path"]
                    / MAILBOX_SUBDIR
                    / CODE_DICT[code]
                ),
            )
            mailboxes.write_headers("hash\n")
            kwargs["mailboxes"] = mailboxes
        merge_func = self.merge_function_dict[code]
        if not self.std_kwargs["quiet"]:
            ProgressBar(dt=SPINNER_UPDATE_PERIOD).register()
        if self.std_kwargs["parallel"]:
            stats_list = (
                self.std_kwargs["bag"]
                .map(merge_func, **kwargs, **extra_kwargs)
                .compute()
            )
        else:
            stats_list = []
            for args in self.std_kwargs["merge_args"]:
                stats_list.append(merge_func(args, **kwargs, **extra_kwargs))
        stats = (
            pd.DataFrame.from_dict(stats_list).set_index("idx").sort_index()
        )
        proteomes = log_and_add_to_stats(proteomes, stats)
        if code in self.merger_kw_dict:
            merger = ExternalMerge(
                file_path_func=mailboxes.path_to_mailbox,
                n_merge=self.std_kwargs["n_proteomes"],
            )
            merger.init("hash")
            merge_counter = AmbiguousMerger(
                start_base=sum(self.n_assigned_list),
                **self.merger_kw_dict[code],
            )
            self.unambig, self.ambig = merger.merge(merge_counter)
            mailboxes.delete()
        self.n_assigned_list.append(len(self.unambig))
        self.last_code = code
        self.pass_name = CODE_DICT[code]
        return proteomes

    def add_ambig_to_total_assigned(self):
        """Include the most recent ambiguous assignment in the total."""
        self.n_assigned_list.append(len(self.ambig))
        self.log_ambig = True

    def get_total_assigned(self):
        """Return the total assigned anchors."""
        return sum(self.n_assigned_list)


def calculate_synteny_hashes(
    args, mailboxes=None, hasher=None, unambig=None, ambig=None
):
    """Calculate synteny hashes for proteins per-genome."""
    idx, dotpath = args
    outpath = dotpath_to_path(dotpath)
    hom = read_tsv_or_parquet(outpath / HOMOLOGY_FILE)
    hom["tmp.nan_group"] = (
        (hom["hom.cluster"].isnull()).astype(int).cumsum() + 1
    ) * (~hom["hom.cluster"].isnull())
    hom.replace(to_replace={"tmp.nan_group": 0}, value=pd.NA, inplace=True)
    hash_name = hasher.hash_name()
    syn_list = []
    if hasher.thorny:  # drop rows
        hom = hom[hom["hom.cluster"].notna()]
    for unused_id_tuple, subframe in hom.groupby(
        by=["frag.id", "tmp.nan_group"]
    ):
        syn_list.append(hasher.calculate(subframe["hom.cluster"]))
    del hom["tmp.nan_group"]
    syn = hom.join(
        pd.concat([df for df in syn_list if df is not None], axis=0)
    )
    del syn_list
    write_tsv_or_parquet(syn, outpath / SYNTENY_FILE, remove_tmp=False)
    syn["tmp.self_count"] = pd.array(
        syn[hash_name].map(syn[hash_name].value_counts()),
        dtype=pd.UInt32Dtype(),
    )
    unique_hashes = (
        syn[[hash_name, "tmp.self_count"]]
        .drop_duplicates(subset=[hash_name])
        .dropna(how="any")
    )
    unique_hashes = unique_hashes.set_index(hash_name).sort_index()
    with mailboxes.locked_open_for_write(idx) as file_handle:
        unique_hashes.to_csv(file_handle, header=False, sep="\t")
    return {
        "idx": idx,
        "path": dotpath,
        "hom.clusters": syn["hom.cluster"].notna().sum(),
        "syn.hashes.n": syn[hash_name].notna().sum(),
    }


def merge_unambig_hashes(
    args,
    unambig=None,
    ambig=None,
    hasher=None,
    mailboxes=None,
):
    """Merge unambiguous synteny hashes into proteomes per-proteome."""
    hash_name = hasher.hash_name()
    idx, dotpath = args
    outpath = dotpath_to_path(dotpath)
    syn = read_tsv_or_parquet(outpath / SYNTENY_FILE)
    syn = _join_on_col_with_na(syn, unambig, hash_name)
    syn = _join_on_col_with_na(syn, ambig, hash_name)
    syn["syn.code"] = pd.NA
    syn["syn.code"] = _fill_col1_val_where_col2_notna(
        syn["syn.code"], syn["syn.anchor.id"], UNAMBIGUOUS_CODE
    )
    # Calculate disambiguation hashes and write them out for merge
    disambig_frame_list = []
    for unused_frag, subframe in syn.groupby(by=["frag.id"]):
        disambig_frame_list.append(hasher.calculate_disambig_hashes(subframe))
    disambig_fr = pd.concat(
        [df for df in disambig_frame_list if df is not None]
    )
    disambig_fr = disambig_fr.dropna(how="all")
    syn = syn.join(disambig_fr)
    write_tsv_or_parquet(syn, outpath / SYNTENY_FILE, remove_tmp=False)
    # Write out unified upstream/downstream hash values
    merged_hashes = pd.concat(
        [
            _rename_and_fill_alt(syn, "tmp.disambig.up", "tmp.disambig.down"),
            _rename_and_fill_alt(syn, "tmp.disambig.down", "tmp.disambig.up"),
        ],
        ignore_index=True,
    )
    merged_hashes["self_count"] = pd.array(
        merged_hashes["hash"].map(merged_hashes["hash"].value_counts()),
        dtype=pd.UInt32Dtype(),
    )
    merged_hashes = merged_hashes.reindex(
        columns=["hash", "self_count", "alt_hash"]
    )
    unique_hashes = (
        merged_hashes.drop_duplicates(subset=["hash"])
        .set_index("hash")
        .sort_index()
    )
    del merged_hashes
    with mailboxes.locked_open_for_write(idx) as file_handle:
        unique_hashes.to_csv(file_handle, header=False, sep="\t")
    return {
        "idx": idx,
        "path": dotpath,
        "syn.anchors.unambiguous": _count_code(
            syn["syn.code"], UNAMBIGUOUS_CODE
        ),
    }


def merge_disambig_hashes(
    args,
    unambig=None,
    ambig=None,
    hasher=None,
    mailboxes=None,
):
    """Merge disambiguated synteny hashes into proteomes per-proteome."""
    idx, dotpath = args
    plain_hash_name = hasher.hash_name(no_prefix=True)
    hash_name = "syn." + plain_hash_name
    outpath = dotpath_to_path(dotpath)
    syn = read_tsv_or_parquet(outpath / SYNTENY_FILE)
    syn = _join_on_col_with_na(syn, unambig, "tmp.disambig.up")
    syn = _join_on_col_with_na(syn, unambig, "tmp.disambig.down")
    for dup_col in [
        "tmp.disambig.anchor.count",
        "tmp.disambig.anchor.id",
    ]:
        xcol = dup_col + "_x"
        ycol = dup_col + "_y"
        syn[dup_col] = syn[xcol].fillna(syn[ycol])
        del syn[xcol], syn[ycol]
    syn["syn.anchor.id"] = syn["syn.anchor.id"].fillna(
        syn["tmp.disambig.anchor.id"]
    )
    syn["syn.anchor.count"] = syn["syn.anchor.count"].fillna(
        syn["tmp.disambig.anchor.count"]
    )
    syn["syn.code"] = _fill_col1_val_where_col2_notna(
        syn["syn.code"], syn["tmp.disambig.anchor.id"], DISAMBIGUATED_CODE
    )
    # Delete some non-needed tmp columns
    non_needed_cols = [
        "tmp.disambig.anchor.count",
        "tmp.disambig.anchor.id",
        "tmp.disambig.up",
        "tmp.disambig.down",
    ]
    syn = syn.drop(columns=non_needed_cols)
    # null hashes are already assigned
    syn[hash_name][syn["syn.anchor.id"].notna()] = pd.NA
    write_tsv_or_parquet(syn, outpath / SYNTENY_FILE, remove_tmp=False)
    # Write out non-ambiguous hashes
    syn["tmp.self_count"] = pd.array(
        syn[hash_name].map(syn[hash_name].value_counts()),
        dtype=pd.UInt32Dtype(),
    )
    unique_hashes = (
        syn[[hash_name, "tmp.self_count"]]
        .drop_duplicates(subset=[hash_name])
        .dropna(how="any")
    )
    unique_hashes = unique_hashes.set_index(hash_name).sort_index()
    with mailboxes.locked_open_for_write(idx) as file_handle:
        unique_hashes.to_csv(file_handle, header=False, sep="\t")
    # logger.debug(f"{dotpath} has {syn['syn.anchor.id'].notna().sum()} assignments")
    return {
        "idx": idx,
        "path": dotpath,
        "syn.anchors.disambiguated": _count_code(
            syn["syn.code"], DISAMBIGUATED_CODE
        ),
    }


def merge_nonambig_hashes(
    args,
    join_mb=None,
    unambig=None,
    ambig=None,
    hasher=None,
    n_proteomes=None,
    cluster_mb=None,
    anchor_mb=None,
    write_ambiguous=True,
):
    """Merge disambiguated synteny hashes into proteomes per-proteome."""
    idx, dotpath = args
    plain_hash_name = hasher.hash_name(no_prefix=True)
    hash_name = "syn." + plain_hash_name
    outpath = dotpath_to_path(dotpath)
    syn = read_tsv_or_parquet(outpath / SYNTENY_FILE)
    syn = _join_on_col_with_na(syn, unambig, hash_name)
    #
    # Do the indirects (formerly ambig made nonambig)
    #
    syn["syn.anchor.id"] = syn["syn.anchor.id"].fillna(
        syn["tmp.nonambig.anchor.id"]
    )
    syn["syn.anchor.count"] = syn["syn.anchor.count"].fillna(
        syn["tmp.nonambig.anchor.count"]
    )
    syn["syn.code"] = _fill_col1_val_where_col2_notna(
        syn["syn.code"], syn["tmp.nonambig.anchor.id"], INDIRECT_CODE
    )
    #
    # Do the nonambig (w.r.t. this proteome) and ambig, if requested
    #
    syn = _join_on_col_with_na(syn, ambig, hash_name)
    n_proteins = len(syn)
    syn["tmp.i"] = range(n_proteins)
    ambig_code = syn["syn.code"].copy()
    ambig_ids = syn["tmp.ambig.anchor.id"].copy()
    ambig_counts = syn["tmp.ambig.anchor.count"].copy()
    for unused_ambig_id, subframe in syn.groupby(by=["tmp.ambig.anchor.id"]):
        ambig_n = len(subframe)
        for unused_i, row in subframe.iterrows():
            row_no = row["tmp.i"]
            if ambig_n == 1:
                ambig_code.iloc[row_no] = LOCALLY_UNAMBIGUOUS_CODE
            elif ambig_n > 1:
                if write_ambiguous:
                    ambig_code.iloc[row_no] = AMBIGUOUS_CODE
                else:
                    ambig_ids.iloc[row_no] = pd.NA
                    ambig_counts.iloc[row_no] = pd.NA
    syn["syn.anchor.id"] = syn["syn.anchor.id"].fillna(ambig_ids)
    syn["syn.anchor.count"] = syn["syn.anchor.count"].fillna(ambig_counts)
    syn["syn.code"] = syn["syn.code"].fillna(ambig_code)
    del ambig_code, ambig_ids, ambig_counts
    #
    # Hsh footprint and direction are anchor properties, where set
    #
    syn = syn.rename(
        columns={
            "syn.hash.footprint": "syn.anchor.footprint",
            "syn.hash.direction": "syn.anchor.direction",
        }
    )
    n_anchors = syn["syn.anchor.id"].notna().sum()  # before shingling
    #
    # Do shingling
    #
    # shingle_id = np.array([np.nan] * n_proteins)
    # shingle_count = np.array([np.nan] * n_proteins)
    # shingle_code = syn["syn.code"].to_numpy()
    # shingle_direction = syn["syn.anchor.direction"].to_numpy()
    # shingle_sub = np.array([np.nan] * n_proteins)
    with join_mb.locked_open_for_write(idx) as file_handle:
        for anchor_count, subframe in syn.groupby(by=["syn.anchor.count"]):
            for unused_i, row in subframe.iterrows():
                anchor_id = row["syn.anchor.id"]
                if pd.isnull(anchor_id):
                    continue
                first_row = row["tmp.i"]
                last_row = first_row + row["syn.anchor.footprint"]
                # shingle_id[first_row:last_row] = anchor_id
                # shingle_count[first_row:last_row] = anchor_count
                # shingle_code[first_row:last_row] = row["syn.code"]
                # shingle_direction[first_row:last_row] = row[
                #    "syn.anchor.direction"
                # ]
                # shingle_sub[first_row:last_row] = hasher.shingle(
                #    syn["hom.cluster"][first_row:last_row],
                #    row["syn.anchor.direction"],
                #    row[hash_name],
                # )
                shingle_fr = pd.DataFrame(
                    {
                        "member_ids": syn.iloc[first_row:last_row].index,
                        "syn.anchor.sub_id": hasher.shingle(
                            syn["hom.cluster"][first_row:last_row],
                            row["syn.anchor.direction"],
                            row[hash_name],
                        ),
                    }
                )
                shingle_fr["syn.anchor.id"] = anchor_id
                shingle_fr["syn.anchor.count"] = anchor_count
                shingle_fr["syn.code"] = row["syn.code"]
                shingle_fr["frag.idx"] = row["frag.idx"]
                shingle_fr.to_csv(file_handle, header=False, sep="\t")
    # syn["syn.anchor.id"] = shingle_id
    # syn["syn.anchor.count"] = shingle_count
    # syn["syn.code"] = shingle_code
    # syn["syn.anchor.direction"] = shingle_direction
    # syn["syn.anchor.sub_id"] = shingle_sub
    # del shingle_id, shingle_count, shingle_code, shingle_sub
    # Delete non-needed (but non-tmp) columns
    # non_needed_cols = [hash_name]
    # syn = syn.drop(columns=non_needed_cols)
    # Write proteome file
    # write_tsv_or_parquet(
    #    syn,
    #    outpath / SYNTENY_FILE,
    # )
    # Write anchor info to mailbox
    # for anchor_id, subframe in syn.groupby(by=["syn.anchor.id"]):
    #    anchor_frame = subframe.copy()
    #    anchor_frame["path"] = dotpath
    #    with anchor_mb.locked_open_for_write(anchor_id) as file_handle:
    #        anchor_frame[ANCHOR_COLS].to_csv(
    #            file_handle, header=False, sep="\t"
    #        )
    # for cluster_id, subframe in syn.groupby(by=["hom.cluster"]):
    #    with cluster_mb.locked_open_for_write(cluster_id) as file_handle:
    #        subframe[CLUSTER_COLS].to_csv(file_handle, header=False, sep="\t")
    in_synteny = syn["syn.anchor.id"].notna().sum()
    n_assigned = syn["hom.cluster"].notna().sum()
    avg_ortho = syn["syn.anchor.count"].mean()
    synteny_pct = in_synteny * 100.0 / n_assigned
    n_ambig = _count_code(syn["syn.code"], AMBIGUOUS_CODE)
    n_nonambig = in_synteny - n_ambig
    nonambig_pct = n_nonambig * 100.0 / n_assigned
    synteny_stats = {
        "idx": idx,
        "path": dotpath,
        "syn.anchors.indirect_unambiguous": _count_code(
            syn["syn.code"], INDIRECT_CODE
        ),
        "syn.anchors.locally_unambiguous": _count_code(
            syn["syn.code"], LOCALLY_UNAMBIGUOUS_CODE
        ),
        "syn.anchors.ambiguous": n_ambig,
        "syn.anchors.nonambiguous": n_nonambig,
        "syn.anchors.nonambig_pct": nonambig_pct,
        "syn.anchors.base": n_anchors,
        "syn.anchors.total": in_synteny,
        "syn.anchors.total_pct": synteny_pct,
        "syn.orthogenomic_pct": avg_ortho * 100.0 / n_proteomes,
    }
    return synteny_stats


def join_synteny_to_clusters(args, cluster_parent=None, mailbox_reader=None):
    """Read homology info from mailbox and join it to proteome file."""
    idx = args[0]
    cluster_path = cluster_parent / f"{idx}.parq"
    cluster = pd.read_parquet(cluster_path)
    n_cluster = len(cluster)
    with mailbox_reader(idx) as file_handle:
        synteny_frame = pd.read_csv(
            file_handle, sep="\t", index_col=0
        ).convert_dtypes()
        in_synteny = len(synteny_frame)
    # cluster files are unusual in that I don't bother to version them,
    # so overlapping info has to be deleted each time
    clust_syn = _concat_without_overlap(cluster, synteny_frame)
    write_tsv_or_parquet(clust_syn, cluster_path)
    anchor_count = clust_syn["syn.anchor.id"].value_counts()
    anchor_frag_counts = [0]
    for unused_id_tuple, subframe in clust_syn.groupby(
        by=["syn.anchor.id", "path"]
    ):
        if len(subframe) == 1:
            anchor_frag_counts.append(1)
        else:
            anchor_frag_counts.append(len(subframe["frag.idx"].value_counts()))
    return {
        "clust_id": idx,
        "in_synteny": in_synteny,
        "n_anchors": len(anchor_count),
        "max_frags_per_anch": max(anchor_frag_counts),
        "synteny_pct": in_synteny * 100.0 / n_cluster,
    }


def write_anchor(args, synteny_parent=None, mailbox_reader=None):
    """Read synteny anchor info from mailbox and join it to synteny file."""
    idx = args[0]
    with mailbox_reader(idx) as file_handle:
        anchor_frame = pd.read_csv(
            file_handle, sep="\t", index_col=0
        ).convert_dtypes()
    in_anchor = len(anchor_frame)
    if in_anchor == 0:
        return None
    # drop any duplicated ID's--normally shouldn't happen
    anchor_frame.drop(
        anchor_frame[anchor_frame.index.duplicated()].index, inplace=True
    )
    anchor_frame.sort_values(
        by=["syn.anchor.sub_id", "frag.idx", "frag.pos"], inplace=True
    )
    # Make a dictionary of common anchor properties, order will be kept
    anchor_props = {
        "anchor.id": idx,
        "sub": None,
        "code": None,
        "count": None,
        "n": None,
        "n_ambig": None,
        "n_adj": None,
        "adj_groups": None,
        "frag.direction": None,
        "syn.anchor.direction": None,
        "anchor.subframe.ok": True,
        "hash": None,
    }
    code_set = set(anchor_frame["syn.code"])
    for test_code in CODE_DICT.keys():
        if test_code in code_set:
            anchor_props["code"] = test_code
            break
    bad_subframe = False
    prop_list = []
    for sub_no, subframe in anchor_frame.groupby(by=["syn.anchor.sub_id"]):
        (subanchor_props, anchor_subframe, bad_subframe) = _subframe_props(
            anchor_props, subframe, sub_no
        )
        if bad_subframe:
            break
        write_tsv_or_parquet(
            anchor_subframe,
            synteny_parent / f"{idx}.{sub_no}.{SYNTENY_FILETYPE}",
            sort_cols=False,
        )
        prop_list.append(subanchor_props)
    if bad_subframe:  # Probably means a hash collision
        logger.error(f"bad anchor set {idx}")
        prop_list = []
        sub_no = 0
        anchor_props["anchor.subframe.ok"] = False
        for cluster_id, subframe in anchor_frame.groupby(by=["hom.cluster"]):
            (
                subanchor_props,
                anchor_subframe,
                unused_bad_subframe,
            ) = _subframe_props(anchor_props, subframe, sub_no)
            write_tsv_or_parquet(
                anchor_subframe,
                synteny_parent / f"{idx}.{sub_no}.{SYNTENY_FILETYPE}",
                sort_cols=False,
            )
            sub_no += 1
            prop_list.append(subanchor_props)
    return prop_list


def _subframe_props(anchor_props, subframe, sub_no):
    """Calculate subframe properties and write subframe"""
    anchor_subframe = subframe.copy()
    subanchor_props = anchor_props.copy()
    subanchor_props["sub"] = sub_no
    anchor_dir_set = set(anchor_subframe["syn.anchor.direction"])
    if len(anchor_dir_set) == 1:
        subanchor_props["syn.anchor.direction"] = list(anchor_dir_set)[0]
    frag_dir_set = set(anchor_subframe["frag.direction"])
    if len(frag_dir_set) == 1:
        subanchor_props["frag.direction"] = list(frag_dir_set)[0]
    subanchor_props["count"] = anchor_subframe["syn.anchor.count"].iloc[0]
    subanchor_props["n_ambig"] = _count_code(
        anchor_subframe["syn.code"], AMBIGUOUS_CODE
    )
    hom_clust_set = set(anchor_subframe["hom.cluster"])
    if len(hom_clust_set) == 1:
        subanchor_props[f"anchor.{sub_no}.cluster"] = list(hom_clust_set)[0]
    else:
        bad_subframe = True
    del (
        anchor_subframe["syn.anchor.count"],
        anchor_subframe["syn.anchor.sub_id"],
    )
    subanchor_props["n"] = len(anchor_subframe)
    subanchor_props["hash"] = hash_array(
        np.sort(anchor_subframe.index.to_numpy())
    )
    (
        subanchor_props["n_adj"],
        subanchor_props["adj_groups"],
        unused_adj_group,
    ) = calculate_adjacency_group(
        anchor_subframe["frag.pos"], anchor_subframe["frag.idx"]
    )
    return subanchor_props, anchor_subframe, bad_subframe


def _concat_without_overlap(df1, df2):
    """Concatenate df2 on top of df1."""
    overlapping = set(df1.columns).intersection(df2.columns)
    if len(overlapping) > 0:
        df1 = df1.drop(columns=overlapping)
    return pd.concat([df1, df2], axis=1)


def _rename_and_fill_alt(df1, key, alt_key):
    """Rename columns and zero-fill alternate."""
    df2 = df1[[key]].rename(columns={key: "hash"})
    df2["alt_hash"] = df1[alt_key].fillna(0)
    return df2.dropna(how="any")


def _join_on_col_with_na(left, right, col_name):
    """Join on a temporary column of type 'O'."""
    tmp_col_name = "tmp." + col_name
    left[tmp_col_name] = left[col_name].astype("O")
    merged = pd.merge(
        left, right, left_on=tmp_col_name, right_index=True, how="left"
    )
    del merged[tmp_col_name]
    return merged


def _fill_col1_val_where_col2_notna(col1, col2, val):
    """Set col1 to val  where col2 is not NA if col1 is not set."""
    fill_ser = col1.copy()
    fill_ser[col2.notna()] = val
    return col1.fillna(fill_ser)


def _count_code(code_ser, code):
    """Counts number of occurrances of code in code_ser."""
    return (code_ser == code).sum()


def anchors_to_adjacency(set_path, n_proteomes, mailbox_reader):
    """Merge adjacencies and produce and adjacency graph."""
    frame_list = []
    for idx in range(n_proteomes):
        with mailbox_reader(idx) as file_handle:
            frame_list.append(
                pd.read_csv(
                    file_handle, sep="\t", index_col=0
                ).convert_dtypes()
            )
    nodes = pd.concat(
        frame_list,
        ignore_index=True,
    )
    del frame_list
    graph = nx.Graph()
    for unused_tuple, subframe in nodes.groupby(
        by=["syn.anchor.id", "syn.anchor.sub_id"]
    ):
        ids = subframe["member_ids"]
        n_ids = len(ids)
        graph.add_nodes_from(ids)
        if n_ids > 1:
            edges = combinations(ids, 2)
            graph.add_edges_from(edges, weight=n_ids)
    outpath = set_path / ANCHORS_FILE
    summarypath = outpath.parent / (
        outpath.name[: -len(outpath.suffix)] + "_summary.tsv"
    )
    histpath = outpath.parent / (
        outpath.name[: -len(outpath.suffix)] + "_hist.tsv"
    )
    components = [
        c
        for c in sorted(nx.connected_components(graph), key=len, reverse=True)
        if len(c) > 1
    ]
    fh = outpath.open("w")
    fh.write("idx\tcluster_id\tsize\tmembers\n")
    n_items = 0
    count_list = []
    hash_list = []
    id_list = []
    for i, comp in enumerate(components):
        component = np.sort(pd.Index(list(comp)).to_numpy())
        id_list.append(i)
        size = len(comp)
        count_list.append(size)
        hash_list.append(hash_array(component))
        for node in component:
            fh.write(f"{n_items}\t{i}\t{size}\t{node}\n")
            n_items += 1
    fh.close()
    n_clusts = len(count_list)
    del graph, components
    cluster_counts = pd.DataFrame({"size": count_list})
    largest_cluster = cluster_counts["size"].max()
    cluster_hist = (
        pd.DataFrame(cluster_counts.value_counts()).sort_index().reset_index()
    )
    cluster_hist = cluster_hist.set_index("size")
    cluster_hist = cluster_hist.rename(columns={0: "n"})
    cluster_hist["item_pct"] = (
        cluster_hist["n"] * cluster_hist.index * 100.0 / n_items
    )
    cluster_hist.to_csv(histpath, sep="\t", float_format="%5.2f")
    cluster_hist["cluster_pct"] = cluster_hist["n"] * 100.0 / n_clusts
    cluster_hist.to_csv(histpath, sep="\t", float_format="%5.2f")
    clusters = pd.DataFrame(
        {"anchor.id": id_list, "count": count_list, "hash": hash_list}
    )
    clusters.to_csv(summarypath, sep="\t")
    stats_dict = {
        "in_anchor": n_items,
        "syn.anchors.n": n_clusts,
        "syn.anchors.largest": largest_cluster,
    }
    return stats_dict


def intersect_anchors(set1_file, set2_file):
    set1_path = Path(set1_file)
    set2_path = Path(set2_file)
    set1_fr = pd.read_csv(set1_path, sep="\t", index_col=0)
    set2_fr = pd.read_csv(set2_path, sep="\t", index_col=0)
    set1_dict = {}
    for cluster_id, subframe in set1_fr.groupby(by=["cluster_id"]):
        set1_dict[cluster_id] = set(subframe["members"].to_numpy())
    set2_dict = {}
    for cluster_id, subframe in set2_fr.groupby(by=["cluster_id"]):
        set2_dict[cluster_id] = set(subframe["members"].to_numpy())
    identity_sets = []
    s1_subset = []
    s2_subset = []
    incongruent = []
    match_keys = list(set2_dict.keys())
    for key1 in set1_dict:
        s1 = set1_dict[key1]
        for i, key2 in enumerate(match_keys):
            s2 = set2_dict[key2]
            if len(s1.intersection(s2)) == 0:
                continue
            elif s1 == s2:
                identity_sets.append(
                    (
                        key1,
                        key2,
                    )
                )
                match_keys.pop(i)
                break
            elif s1.issubset(s2):
                s1_subset.append(
                    (
                        key1,
                        key2,
                    )
                )
                match_keys.pop(i)
                break
            elif s2.issubset(s1):
                s2_subset.append(
                    (
                        key1,
                        key2,
                    )
                )
                match_keys.pop(i)
                break
            else:
                incongruent.append(
                    (
                        key1,
                        key2,
                    )
                )
    logger.info(f"set 1 ({set1_file}): {len(set1_dict)}")
    logger.info(f"set 2 ({set2_file}): {len(set2_dict)}")
    min_sets = min(len(set1_dict), len(set2_dict))
    id_len = len(identity_sets)
    id_pct = id_len * 100.0 / min_sets
    logger.info(f"identity: {id_len} ({id_pct:.1f}%)")
    s1_len = len(s1_subset)
    s1_pct = s1_len * 100.0 / min_sets
    logger.info(f"set 1 is subset: {s1_len} ({s1_pct:.1f}%)")
    s2_len = len(s2_subset)
    s2_pct = s2_len * 100.0 / min_sets
    logger.info(f"set 2 is subset: {s2_len} ({s2_pct:.1f}%)")
    incon_len = len(incongruent)
    incon_pct = incon_len * 100.0 / min_sets
    logger.info(f"incongruent: {incon_len}({incon_pct}%)")



# -*- coding: utf-8 -*-
"""Homology (sequence similarity) operations."""
# standard library imports
import fcntl
import json
import os
import shutil
import sys
from pathlib import Path

# third-party imports
import dask.bag as db
import pandas as pd
from dask.diagnostics import ProgressBar

# first-party imports
import sh

# module imports
from .common import CLUSTER_FILETYPE
from .common import CLUSTERS_FILE
from .common import EXTERNAL_CLUSTERS_FILE
from .common import FRAGMENTS_FILE
from .common import HOMOLOGY_FILE
from .common import PROTEINS_FILE
from .common import PROTEOMES_FILE
from .common import PROTEOMOLOGY_FILE
from .common import SEARCH_PATHS
from .common import SPINNER_UPDATE_PERIOD
from .common import TrimmableMemoryMap
from .common import calculate_adjacency_group
from .common import dotpath_to_path
from .common import group_key_filename
from .common import logger
from .common import read_tsv_or_parquet
from .common import sort_proteome_frame
from .common import write_tsv_or_parquet
from .core import homology_cluster
from .mailboxes import DataMailboxes

# global constants
HOMOLOGY_COLS = ["hom.cluster", "hom.cl_size"]


def cluster_build_trees(
    identity, set_name, cluster_file=None, click_loguru=None
):
    """Calculate homology clusters, MSAs, trees."""
    options = click_loguru.get_global_options()
    user_options = click_loguru.get_user_global_options()
    parallel = user_options["parallel"]
    set_path = Path(set_name)
    # read and possibly update proteomes
    proteomes_path = set_path / PROTEOMES_FILE
    proteomes_in = read_tsv_or_parquet(proteomes_path)
    proteomes = sort_proteome_frame(proteomes_in)
    if not proteomes_in.equals(proteomes):
        logger.info("proteomes sort order changed, writing new proteomes file")
        write_tsv_or_parquet(proteomes, proteomes_path)
    n_proteomes = len(proteomes)
    # read and update fragment ID's
    frags = read_tsv_or_parquet(set_path / FRAGMENTS_FILE)
    frags["frag.idx"] = pd.array(frags.index, dtype=pd.UInt32Dtype())
    frag_frames = {}
    for dotpath, subframe in frags.groupby(by=["path"]):
        frag_frames[dotpath] = subframe.copy().set_index("frag.orig_id")
    arg_list = []
    concat_fasta_path = set_path / "proteins.fa"
    for i, row in proteomes.iterrows():
        arg_list.append((row, concat_fasta_path, frag_frames[row["path"]]))
    file_idx = {}
    stem_dict = {}
    for i, row in proteomes.iterrows():
        stem = row["path"]
        file_idx[stem] = i
        stem_dict[i] = stem
    if cluster_file is None:
        if concat_fasta_path.exists():
            concat_fasta_path.unlink()
        if not options.quiet:
            logger.info(
                f"Renaming fragments and concatenating sequences for {len(arg_list)}"
                " proteomes:"
            )
        for args in arg_list:
            write_protein_fasta(args)
        del arg_list
        cwd = Path.cwd()
        os.chdir(set_path)
        n_clusters, run_stats, cluster_hist = homology_cluster(
            "proteins.fa",
            identity,
            write_ids=True,
            delete=False,
            cluster_stats=False,
            outname="homology",
            click_loguru=click_loguru,
        )
        log_path = Path("homology.log")
        log_dir_path = Path("logs")
        log_dir_path.mkdir(exist_ok=True)
        shutil.copy2(log_path, "logs/homology.log")
        log_path.unlink()
        os.chdir(cwd)
        logger.info(f"Number of clusters: {n_clusters}")
        del cluster_hist
        del run_stats
        concat_fasta_path.unlink()
    else:  # use pre-existing clusters
        homology_path = set_path / "homology"
        if homology_path.exists():
            shutil.rmtree(homology_path)
        inclusts = pd.read_csv(cluster_file, sep="\t")
        for col in ["cluster_id", "members"]:
            if col not in inclusts.columns:
                logger.error(
                    f'Column named "{col}" not found in external homology cluster file'
                )
                sys.exit(1)
        cluster_counts = inclusts["cluster_id"].value_counts()
        cluster_map = pd.Series(
            range(len(cluster_counts)), index=cluster_counts.index
        )
        cluster_ids = inclusts["cluster_id"].map(cluster_map)
        cluster_sizes = inclusts["cluster_id"].map(cluster_counts)
        predef_clusters = pd.DataFrame(
            {
                "cluster_id": cluster_ids,
                "size": cluster_sizes,
                "members": inclusts["members"],
            }
        )
        predef_clusters.sort_values(by=["cluster_id"], inplace=True)
        predef_clusters.drop(
            predef_clusters[predef_clusters["size"] < 2].index,
            axis=0,
            inplace=True,
        )
        n_clusters = predef_clusters["cluster_id"].max() + 1
        predef_clusters.index = range(len(predef_clusters))
        external_cluster_path = set_path / EXTERNAL_CLUSTERS_FILE
        logger.info(
            f"Writing {external_cluster_path} with {len(predef_clusters)} genes"
            + f" in {n_clusters} homology clusters"
        )
        predef_clusters.to_csv(external_cluster_path, sep="\t")
        del cluster_counts, cluster_map, cluster_sizes, inclusts
        homology_path = set_path / "homology"
        homology_path.mkdir(exist_ok=True)
        if not options.quiet:
            logger.info(
                f"Creating cluster files for for {len(arg_list)}" " proteomes:"
            )
        proteome_no = 0
        for args in arg_list:
            logger.info(f"doing proteome {proteome_no}")
            write_protein_fasta(
                args, fasta_dir=homology_path, clusters=predef_clusters
            )
            proteome_no += 1
        del arg_list
        logger.info(
            "Checking that all cluster files are present (gene-id mismatch)"
        )
        missing_files = False
        for i in range(n_clusters):
            if not (homology_path / f"{i}.fa").exists():
                logger.error(f"External cluster {i} is missing.")
                missing_files = True
        if missing_files:
            sys.exit(1)
    #
    # Write homology info back into proteomes
    #
    click_loguru.elapsed_time("Alignment/tree-building")
    hom_mb = DataMailboxes(
        n_boxes=n_proteomes,
        mb_dir_path=(set_path / "mailboxes" / "clusters2proteomes"),
        file_extension="tsv",
    )
    hom_mb.write_tsv_headers(HOMOLOGY_COLS)
    cluster_paths = [
        set_path / "homology" / f"{i}.fa" for i in range(n_clusters)
    ]
    bag = db.from_sequence(cluster_paths)
    cluster_stats = []
    if not options.quiet:
        logger.info(
            f"Calculating MSAs and trees for {len(cluster_paths)} homology"
            " clusters:"
        )
        ProgressBar(dt=SPINNER_UPDATE_PERIOD).register()
    if parallel:
        cluster_stats = bag.map(
            parse_cluster,
            file_dict=file_idx,
            file_writer=hom_mb.locked_open_for_write,
        )
    else:
        for clust_fasta in cluster_paths:
            cluster_stats.append(
                parse_cluster(
                    clust_fasta,
                    file_dict=file_idx,
                    file_writer=hom_mb.locked_open_for_write,
                )
            )
    n_clust_genes = 0
    clusters_dict = {}
    for cluster_id, cluster_dict in cluster_stats:
        n_clust_genes += cluster_dict["size"]
        clusters_dict[cluster_id] = cluster_dict
    del cluster_stats
    clusters = pd.DataFrame.from_dict(clusters_dict).transpose()
    del clusters_dict
    clusters.sort_index(inplace=True)
    grouping_dict = {}
    for i in range(n_proteomes):  # keep numbering of single-file clusters
        grouping_dict[f"[{i}]"] = i
    grouping_dict[str(list(range(n_proteomes)))] = 0
    for n_members, subframe in clusters.groupby(["n_memb"]):
        if n_members == 1:
            continue
        if n_members == n_proteomes:
            continue
        member_counts = pd.DataFrame(subframe["n_members"].value_counts())
        member_counts["key"] = range(len(member_counts))
        for newcol in range(n_members):
            member_counts[f"memb{newcol}"] = ""
        for member_string, row in member_counts.iterrows():
            grouping_dict[member_string] = row["key"]
            member_list = json.loads(member_string)
            for col in range(n_members):
                member_counts.loc[member_string, f"memb{col}"] = stem_dict[
                    member_list[col]
                ]
        member_counts = member_counts.set_index("key")
        write_tsv_or_parquet(
            member_counts, set_path / group_key_filename(n_members)
        )
    clusters["n_members"] = clusters["n_members"].map(grouping_dict)
    clusters = clusters.rename(columns={"n_members": "group_key"})
    n_adj = clusters["n_adj"].sum()
    adj_pct = n_adj * 100.0 / n_clust_genes
    n_adj_clust = sum(clusters["adj_groups"] != 0)
    adj_clust_pct = n_adj_clust * 100.0 / len(clusters)
    logger.info(
        f"{n_adj} ({adj_pct:.1f}%) out of {n_clust_genes}"
        + " clustered genes are adjacent"
    )
    logger.info(
        f"{n_adj_clust} ({adj_clust_pct:.1f}%) out of "
        + f"{len(clusters)} clusters contain adjacency"
    )
    write_tsv_or_parquet(clusters, set_path / CLUSTERS_FILE)
    # join homology cluster info to proteome info
    click_loguru.elapsed_time("Joining")
    arg_list = []
    for i, row in proteomes.iterrows():
        arg_list.append(
            (
                i,
                dotpath_to_path(row["path"]),
            )
        )
    bag = db.from_sequence(arg_list)
    hom_stats = []
    if not options.quiet:
        logger.info(f"Joining homology info to {n_proteomes} proteomes:")
        ProgressBar(dt=SPINNER_UPDATE_PERIOD).register()
    if parallel:
        hom_stats = bag.map(
            join_homology_to_proteome, mailbox_reader=hom_mb.open_then_delete
        ).compute()
    else:
        for args in arg_list:
            hom_stats.append(
                join_homology_to_proteome(
                    args, mailbox_reader=hom_mb.open_then_delete
                )
            )
    hom_mb.delete()
    hom_frame = pd.DataFrame.from_dict(hom_stats)
    hom_frame.set_index(["prot.idx"], inplace=True)
    hom_frame.sort_index(inplace=True)
    logger.info("Homology cluster coverage:")
    with pd.option_context(
        "display.max_rows", None, "display.float_format", "{:,.2f}%".format
    ):
        logger.info(hom_frame)
    proteomes = pd.concat([proteomes, hom_frame], axis=1)
    write_tsv_or_parquet(
        proteomes, set_path / PROTEOMOLOGY_FILE, float_format="%5.2f"
    )
    click_loguru.elapsed_time(None)


def write_protein_fasta(args, clusters=None, fasta_dir=None):
    """Read peptide sequences from info file and write them out."""
    row, concat_fasta_path, frags = args
    dotpath = row["path"]
    phylogeny_dict = {"prot.idx": row.name, "path": dotpath}
    for phy_prop in [name for name in row.index if name.startswith("phy.")]:
        phylogeny_dict[phy_prop] = row[phy_prop]
    inpath = dotpath_to_path(dotpath)
    prot_info = read_tsv_or_parquet(inpath / PROTEINS_FILE)
    prot_info["frag.idx"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.idx"]
    )
    prot_info["frag.is_plas"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_plas"]
    )
    prot_info["frag.is_scaf"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_scaf"]
    )
    prot_info["frag.is_chr"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.is_chr"]
    )
    prot_info["frag.id"] = prot_info["frag.id"].map(
        lambda oid: frags.loc[oid]["frag.id"]
    )
    # Write out updated protein info
    write_tsv_or_parquet(prot_info, inpath / HOMOLOGY_FILE)
    # include phylogeny info in per-sequence info
    for prop in phylogeny_dict:
        prot_info[prop] = phylogeny_dict[prop]
    # write concatenated sequence info
    if clusters is None:
        fasta_path = concat_fasta_path
        info_to_fasta(None, fasta_path, append=True, infoobj=prot_info)
    else:
        for cluster_id, subframe in clusters.groupby(by=["cluster_id"]):
            cluster_info = prot_info[prot_info.index.isin(subframe["members"])]
            fasta_path = fasta_dir / f"{cluster_id}.fa"
            info_to_fasta(None, fasta_path, append=True, infoobj=cluster_info)


def parse_cluster(
    fasta_path, file_dict=None, file_writer=None, neighbor_joining=False
):
    """Parse cluster FASTA headers to create cluster table.."""
    cluster_id = fasta_path.name[:-3]
    outdir = fasta_path.parent
    clusters = parse_cluster_fasta(fasta_path)
    if len(clusters) < 2:
        # fasta_path.unlink()
        logger.error(f"Singleton Cluster {cluster_id} is size {len(clusters)}")
        cluster_dict = {
            "size": len(clusters),
            "n_memb": None,
            "n_members": None,
            "n_adj": None,
            "adj_groups": None,
        }
        return int(cluster_id)
    # calculate MSA and return guide tree
    muscle_args = [
        "-in",
        f"{outdir}/{cluster_id}.fa",
        "-out",
        f"{outdir}/{cluster_id}.faa",
        "-diags",
        "-sv",
        "-maxiters",
        "2",
        "-quiet",
        "-distance1",
        "kmer20_4",
    ]
    if len(clusters) >= 4:
        muscle_args += [
            "-tree2",
            f"{outdir}/{cluster_id}.nwk",
        ]
        if neighbor_joining:
            muscle_args += ["-cluster2", "neighborjoining"]  # adds 20%
    try:
        muscle = sh.Command("muscle", search_paths=SEARCH_PATHS)
    except sh.CommandNotFound:
        logger.error("muscle must be installed first.")
        sys.exit(1)
    muscle(muscle_args)
    # fasta_path.unlink()
    clusters["prot.idx"] = clusters["path"].map(file_dict)
    clusters.sort_values(by=["prot.idx", "frag.id", "frag.pos"], inplace=True)
    n_adj, adj_gr_count, unused_adj_group = calculate_adjacency_group(
        clusters["frag.pos"], clusters["frag.idx"]
    )
    idx_values = clusters["prot.idx"].value_counts()
    idx_list = list(idx_values.index)
    idx_list.sort()
    write_tsv_or_parquet(clusters, outdir / f"{cluster_id}.{CLUSTER_FILETYPE}")
    cluster_dict = {
        "size": len(clusters),
        "n_memb": len(idx_values),
        "n_members": str(idx_list),
        "n_adj": n_adj,
        "adj_groups": adj_gr_count,
    }
    for group_id, subframe in clusters.groupby(by=["prot.idx"]):
        proteome_frame = subframe.copy()
        proteome_frame["hom.cluster"] = cluster_id
        proteome_frame["hom.cl_size"] = len(idx_values)
        proteome_frame.drop(
            proteome_frame.columns.drop(HOMOLOGY_COLS),  # drop EXCEPT these
            axis=1,
            inplace=True,
        )
        with file_writer(group_id) as file_handle:
            proteome_frame.to_csv(file_handle, header=False, sep="\t")
    return int(cluster_id), cluster_dict


def parse_cluster_fasta(filepath, trim_dict=True):
    """Return FASTA headers as a dictionary of properties."""
    next_pos = 0
    properties_dict = {}
    memory_map = TrimmableMemoryMap(filepath)
    with memory_map.map() as mem_map:
        size = memory_map.size
        next_pos = mem_map.find(b">", next_pos)
        while next_pos != -1 and next_pos < size:
            eol_pos = mem_map.find(b"\n", next_pos)
            if eol_pos == -1:
                break
            space_pos = mem_map.find(b" ", next_pos + 1, eol_pos)
            if space_pos == -1:
                raise ValueError(
                    f"Header format is bad in {filepath} header"
                    f" {len(properties_dict)+1}"
                )
            cluster_id = mem_map[next_pos + 1 : space_pos].decode("utf-8")
            payload = json.loads(mem_map[space_pos + 1 : eol_pos])
            properties_dict[cluster_id] = payload
            if trim_dict:
                size = memory_map.trim(space_pos, eol_pos)
            next_pos = mem_map.find(b">", space_pos)
    cluster = (
        pd.DataFrame.from_dict(properties_dict).transpose().convert_dtypes()
    )
    return cluster


def join_homology_to_proteome(args, mailbox_reader=None):
    """Read homology info from mailbox and join it to proteome file."""
    idx, protein_parent = args
    proteins = pd.read_parquet(protein_parent / HOMOLOGY_FILE)
    n_proteins = len(proteins)
    with mailbox_reader(idx) as file_handle:
        homology_frame = pd.read_csv(
            file_handle, sep="\t", index_col=0
        ).convert_dtypes()
        clusters_in_proteome = len(homology_frame)
    proteome_frame = pd.concat([proteins, homology_frame], axis=1)
    write_tsv_or_parquet(proteome_frame, protein_parent / HOMOLOGY_FILE)
    return {
        "prot.idx": idx,
        "hom.clusters": clusters_in_proteome,
        "hom.cluster_pct": clusters_in_proteome * 100.0 / n_proteins,
    }


def info_to_fasta(infofile, fastafile, append, infoobj=None):
    """Convert infofile to FASTA file."""
    if infoobj is None:
        infoobj = read_tsv_or_parquet(infofile)
    if append:
        filemode = "a+"
    else:
        filemode = "w"
    with Path(fastafile).open(filemode) as file_handle:
        fcntl.flock(file_handle, fcntl.LOCK_EX)
        logger.debug(f"Writing to {fastafile} with mode {filemode}.")
        seqs = infoobj["prot.seq"].copy()
        del infoobj["prot.seq"]
        for gene_id, row in infoobj.iterrows():
            file_handle.write(f">{gene_id} {row.to_json()}\n")
            file_handle.write(f"{seqs[gene_id]}\n")
        fcntl.flock(file_handle, fcntl.LOCK_UN)



# -*- coding: utf-8 -*-
"""azulejo -- tile phylogenetic space with subtrees."""
# standard library imports
import locale
import sys
import warnings
from pkg_resources import iter_entry_points

# third-party imports
import click
import sh
from click_plugins import with_plugins
from click_loguru import ClickLoguru

# module imports
from .common import INSTALL_PATH
from .common import SEARCH_PATHS
from .common import NAME
from .common import logger
from .core import homology_cluster as undeco_homology_cluster
from .core import cluster_in_steps as undeco_cluster_in_steps
from .homology import cluster_build_trees as undeco_cluster_build_trees
from .homology import info_to_fasta as undeco_into_to_fasta
from .ingest import find_files as undeco_find_files
from .ingest import ingest_sequences as undeco_ingest_sequences
from .installer import DependencyInstaller
from .parquet import parquet_to_tsv as undeco_parquet_to_tsv
from .proxy import calculate_proxy_genes as undeco_calculate_proxy_genes
from .synteny import intersect_anchors as undeco_intersect_anchors
from .synteny import synteny_anchors as undeco_synteny_anchors
from .taxonomy import print_taxonomic_ranks
from .taxonomy import rankname_to_number

# global constants
LOG_FILE_RETENTION = 3
__version__ = "0.10.0"
MUSCLE_VER = "3.8.1551"
USEARCH_VER = "11.0.667"
DAGCHAINER_TOOL_VER = "1.2"
BLAST_MINOR = "2.11"
BLAST_VER = f"{BLAST_MINOR}.0"
BLAST_PATCHED = f"{BLAST_MINOR}.99"
MMSEQS_VER = "12.0"
MMSEQS_HASH = "113e3212c137d026e297c7540e1fcd039f6812b1"
MMSEQS_SHORT_HASH = MMSEQS_HASH[:4]
DEPENDENCY_DICT = {
    "muscle": {
        "binaries": ["muscle"],
        "tarball": (
            "https://www.drive5.com/muscle/muscle_src_"
            + MUSCLE_VER
            + ".tar.gz"
        ),
        "dir": ".",
        "patches": ["infinity_fix.patch"],
        "version": MUSCLE_VER + ".99",
        "version_command": ["-version"],
        "version_parser": lambda v: v.split()[1],
        "make": ["muscle-pgo"],
        "copy_binaries": ["muscle"],
        "license": "public domain",
        "license_restrictive": False,
    },
    "usearch": {
        "binaries": ["usearch"],
        "dir": ".",
        "version": USEARCH_VER,
        "required": True,
        "version_command": ["-version"],
        "version_parser": lambda v: v.split()[1].split("_")[0],
        "download_binaries": {
            "linux": "https://www.drive5.com/downloads/usearch11.0.667_i86linux32.gz",
            "macos": "https://www.drive5.com/downloads/usearch11.0.667_i86osx32.gz",
        },
        "make": ["usearch"],
        "copy_binaries": ["usearch"],
        "license": "proprietary non-commercial",
        "license_restrictive": True,
        "license_file": "LICENSE.txt",
    },
    "dagchainer-tool": {
        "binaries": ["dagchainer-tool.sh"],
        "tarball": (
            "https://sourceforge.net/projects/dagchainer/files/dagchainer/"
            + "DAGchainer-r02062008/DAGchainer_r02-06-2008.tar.gz/download"
        ),
        "dir": "DAGCHAINER",
        "required": False,
        "version": DAGCHAINER_TOOL_VER,
        "version_command": ["version"],
        "patches": ["dagchainer-0206008.patch"],
        "make": [],
        "copy_binaries": [
            "dagchainer",
            "dagchainer-tool.sh",
            "hash_into_fasta_id.pl",
            "pairs_to_adjacency.py",
            "top_blast_hit.awk",
            "run_DAG_chainer.pl",
        ],
        "license": "GPL-2",
        "license_restrictive": False,
    },
    "blast-longids": {
        "binaries": ["blastn"],
        "tarball": (
            "https://ftp.ncbi.nih.gov/blast/executables/blast+/"
            + BLAST_VER
            + "/ncbi-blast-"
            + BLAST_VER
            + "+-src.tar.gz"
        ),
        "dir": ("ncbi-blast-" + BLAST_VER + "+-src/c++"),
        "required": False,
        "version_command": ["-version"],
        "version": BLAST_PATCHED,
        "version_parser": lambda v: v.split()[1].split("_")[0].rstrip("+"),
        "patches": ["blast_longer_ids.patch", "configure_settings.patch"],
        "configure": [],
        "make": [],
        "make_install": [],
        "license": "public domain",
        "license_restrictive": False,
    },
    "mmseqs": {
        "binaries": ["mmseqs"],
        "tarball": (
            f"https://github.com/soedinglab/MMseqs2/archive/{MMSEQS_SHORT_HASH}.tar.gz"
        ),
        "dir": (f"MMseqs2-{MMSEQS_HASH}"),
        "required": True,
        "version_command": ["version"],
        "version": MMSEQS_VER,
        "version_parser": lambda v: v.strip(),
        "cmake": [
            "-C",
            "mmseqs-azulejo-config.cmake",
            f"-DVERSION_OVERRIDE={MMSEQS_VER}",
        ],
        "make": [],
        "make_install": [],
        "license": "GPL-3",
        "license_restrictive": False,
    },
}
DEFAULT_K = 2
DEFAULT_STEPS = 16

# set locale so grouping works
for localename in ["en_US", "en_US.utf8", "English_United_States"]:
    try:
        locale.setlocale(locale.LC_ALL, localename)
        break
    except locale.Error:
        continue

# set up logging
click_loguru = ClickLoguru(
    NAME, __version__, retention=LOG_FILE_RETENTION, timer_log_level="info"
)


@with_plugins(iter_entry_points(NAME + ".cli_plugins"))
@click_loguru.logging_options
@click.group()
@click_loguru.stash_subcommand()
@click.option(
    "-e",
    "--warnings_as_errors",
    is_flag=True,
    show_default=True,
    default=False,
    help="Treat warnings as fatal.",
    callback=click_loguru.user_global_options_callback,
)
@click.option(
    "--parallel/--no-parallel",
    is_flag=True,
    default=True,
    show_default=True,
    help="Process in parallel.",
    callback=click_loguru.user_global_options_callback,
)
@click.version_option(version=__version__, prog_name=NAME)
def cli(warnings_as_errors, parallel, **kwargs):
    """Azulejo -- tiling genes in subtrees across phylogenetic space.

    \b
    For more information, see the homepage at https://github.com/legumeinfo/azulejo

    Written by Joel Berendzen <joelb@ncgr.org>.
    Copyright (C) 2020. National Center for Genome Resources. All rights reserved.
    License: BSD-3-Clause
    """
    options = click_loguru.get_global_options()
    if warnings_as_errors:
        if not options.quiet:
            print(
                "Runtime warnings (e.g., from pandas) will cause exceptions!"
            )
        warnings.filterwarnings("error")
    f"{parallel}{kwargs}"


@cli.command()
@click.option(
    "--force",
    "-f",
    help="Force overwrites of existing binaries.",
    is_flag=True,
    default=False,
)
@click.option(
    "--accept_licenses",
    "-y",
    help="Accept all licenses.",
    is_flag=True,
    default=False,
)
@click.argument("dependencies", nargs=-1)
@click_loguru.init_logger(logfile=False)
def install(dependencies, force, accept_licenses):
    """
    Check for/install binary dependencies.

    \b
    Example:
        azulejo install all

    """
    options = click_loguru.get_global_options()
    dep_installer = DependencyInstaller(
        DEPENDENCY_DICT,
        pkg_name=NAME,
        install_path=INSTALL_PATH,
        force=force,
        accept_licenses=accept_licenses,
        quiet=options.quiet,
    )
    if dependencies == ():
        print(dep_installer.status(exe_paths=True), end="")
    else:
        dep_installer.install_list(dependencies)


@cli.command()
@click_loguru.init_logger()
@click_loguru.log_elapsed_time(level="info")
@click_loguru.log_peak_memory_use(level="info")
@click.option(
    "--identity",
    "-i",
    default=0.0,
    help="Minimum sequence ID (0-1). [default: lowest]",
)
@click.option(
    "--cluster_file",
    "-c",
    type=click.Path(exists=True),
    default=None,
    show_default=True,
    help="Use pre-existing homology clusters.",
)
@click.argument("setname")
def homology(identity, setname, cluster_file):
    """
    Calculate homology clusters, MSAs, trees.

    \b
    Example:
        azulejo homology glycines

    """
    undeco_cluster_build_trees(
        identity, setname, cluster_file=cluster_file, click_loguru=click_loguru
    )


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--append/--no-append",
    "-a/-x",
    is_flag=True,
    default=True,
    help="Append to FASTA file.",
    show_default=True,
)
@click.argument("parquetfile")
@click.argument("fastafile")
def parquet_to_fasta(parquetfile, fastafile, append):
    """Convert Parquet sequence info to FASTA file."""
    undeco_into_to_fasta(parquetfile, fastafile, append)


@cli.command()
@click_loguru.init_logger()
@click_loguru.log_elapsed_time(level="info")
@click_loguru.log_peak_memory_use(level="info")
@click.option(
    "-k", default=DEFAULT_K, help="Synteny anchor length.", show_default=True
)
@click.option(
    "--peatmer/--kmer",
    default=True,
    is_flag=True,
    show_default=True,
    help="Allow repeats in anchor.",
)
@click.option(
    "--thorny/--no-thorny",
    default=True,
    is_flag=True,
    show_default=True,
    help="Ignore singletons.",
)
@click.option(
    "--write_ambiguous/--no-write-ambiguous",
    default=False,
    is_flag=True,
    show_default=True,
    help="Include ambiguous anchors.",
)
@click.option(
    "--disambig_adj_only/--disambig_all",
    default=True,
    is_flag=True,
    show_default=True,
    help="Disambig only anchors adj to unambig.",
)
@click.argument("setname")
def synteny(k, peatmer, setname, write_ambiguous, thorny, disambig_adj_only):
    """Calculate synteny anchors.

    \b
    Example:
        azulejo synteny glycines

    """
    undeco_synteny_anchors(
        k,
        peatmer,
        setname,
        click_loguru=click_loguru,
        write_ambiguous=write_ambiguous,
        thorny=thorny,
        disambig_adj_only=disambig_adj_only,
    )


@cli.command()
@click_loguru.init_logger()
@click_loguru.log_elapsed_time(level="info")
@click.argument("setname")
@click.argument("compfile")
def intersect_anchors(setname, compfile):
    """Intersect two sets of synteny anchors.

    \b
    Example:
        azulejo intersect_anchors glycine7/synteny_anchors.tsv dagchainer_tool_out/synteny_anchors.tsv

    """
    undeco_intersect_anchors(setname, compfile)


@cli.command()
@click.argument("rankname", nargs=-1)
def taxonomy(rankname):
    """Check/show taxonomic ranks."""
    if rankname == ():
        print_taxonomic_ranks()
    else:
        rankname = rankname[0]
        try:
            rankval = rankname_to_number(rankname)
        except ValueError as error_msg:
            print(f"ERROR: {error_msg}")
            sys.exit(1)
        print(rankval)


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--columns",
    default=False,
    is_flag=True,
    show_default=False,
    help="Print names/dtypes of columns.",
)
@click.option(
    "--index_name",
    default=False,
    is_flag=True,
    show_default=False,
    help="Print the name of the index.",
)
@click.option(
    "--no_index",
    default=False,
    is_flag=True,
    show_default=False,
    help="Do not write index column.",
)
@click.option(
    "--pretty",
    default=False,
    is_flag=True,
    show_default=False,
    help="Pretty-print output.",
)
@click.option(
    "--no_header",
    "-h",
    default=False,
    is_flag=True,
    show_default=False,
    help="Do not write the header.",
)
@click.option(
    "--col",
    "-c",
    default=None,
    multiple=True,
    show_default=True,
    help="Write only the named column.",
)
@click.option(
    "--writefile",
    "-w",
    default=False,
    is_flag=True,
    show_default=True,
    help="Write to a TSV file.",
)
@click.option(
    "--index_val",
    "-i",
    default=None,
    multiple=True,
    show_default=True,
    help="Write only the row with this index value.",
)
@click.option(
    "--head",
    default=0,
    show_default=False,
    help="Write only the first N rows.",
)
@click.option(
    "--max_rows",
    default=None,
    show_default=False,
    help="Pretty-print N rows.",
)
@click.option(
    "--max_cols",
    default=None,
    show_default=False,
    help="Pretty-print N cols.",
)
@click.option(
    "--tail",
    default=0,
    show_default=False,
    help="Write only the last N rows.",
)
@click.argument("parquetfile", type=click.Path(exists=True))
@click.argument("tsvfile", nargs=-1)
def parquet_to_tsv(
    parquetfile,
    tsvfile,
    columns,
    index_name,
    col,
    no_index,
    head,
    tail,
    index_val,
    no_header,
    pretty,
    max_rows,
    max_cols,
    writefile,
):
    """
    Reads parquet file, writes tsv.

    \b
    Example:
        azulejo parquet-to-tsv glycines/proteomes.syn.parq
    """
    undeco_parquet_to_tsv(
        parquetfile,
        tsvfile,
        columns,
        index_name,
        col,
        no_index,
        head,
        tail,
        index_val,
        no_header,
        pretty,
        max_rows,
        max_cols,
        writefile,
    )


@cli.command()
@click_loguru.init_logger()
@click.argument("seqfile")
@click.option(
    "--identity",
    "-i",
    default=0.0,
    help="Minimum sequence identity (float, 0-1). [default: lowest]",
)
@click.option(
    "--min_id_freq",
    "-m",
    default=0,
    show_default=True,
    help="Minimum frequency of ID components.",
)
@click.option(
    "--delete/--no-delete",
    "-d/-n",
    is_flag=True,
    default=True,
    help="Delete primary output of clustering. [default: delete]",
)
@click.option(
    "--write_ids/--no-write_ids",
    "-w",
    is_flag=True,
    default=False,
    help="Write file of ID-to-clusters. [default: delete]",
)
@click.option(
    "--do_calc/--no-do_calc",
    "-c/-x",
    is_flag=True,
    default=True,
    help="Write file of ID-to-clusters. [default: do_calc]",
)
@click.option(
    "--substrs", help="subpath to file of substrings. [default: none]"
)
@click.option("--dups", help="subpath to file of duplicates. [default: none]")
def cluster(
    seqfile,
    identity,
    delete=True,
    write_ids=False,
    do_calc=True,
    min_id_freq=0,
    substrs=None,
    dups=None,
    cluster_stats=True,
    outname=None,
):
    """Cluster at a global sequence identity threshold."""
    undeco_homology_cluster(
        seqfile,
        identity,
        delete=delete,
        write_ids=write_ids,
        do_calc=do_calc,
        min_id_freq=min_id_freq,
        substrs=substrs,
        dups=dups,
        cluster_stats=cluster_stats,
        outname=outname,
        click_loguru=click_loguru,
    )


@cli.command()
@click_loguru.init_logger()
@click.argument("seqfile")
@click.option(
    "--steps",
    "-s",
    default=DEFAULT_STEPS,
    show_default=True,
    help="# of steps from lowest to highest",
)
@click.option(
    "--min_id_freq",
    "-m",
    default=0,
    show_default=True,
    help="Minimum frequency of ID components.",
)
@click.option(
    "--substrs", help="subpath to file of substrings. [default: none]"
)
@click.option("--dups", help="subpath to file of duplicates. [default: none]")
def cluster_in_steps(seqfile, steps, min_id_freq=0, substrs=None, dups=None):
    """Cluster in steps from low to 100% identity."""
    undeco_cluster_in_steps(
        seqfile, steps, min_id_freq=min_id_freq, substrs=substrs, dups=dups
    )


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.option(
    "--gff_pattern", "-g", multiple=True, help="Glob pattern for GFF"
)
@click.option(
    "--faa_pattern", "-a", multiple=True, help="Glob for protein FASTA"
)
@click.option(
    "--fna_pattern", "-n", multiple=True, help="Glob for nucleic FASTA"
)
@click.option("--exclude", "-e", multiple=True, help="Glob for exclude")
@click.option(
    "--preference", "-o", multiple=True, help="Glob for preference order"
)
@click.option("--name", "-n", multiple=True, help="Name, one per match")
@click.option(
    "--name_from_part",
    type=int,
    default=None,
    help="Path component for name parsing",
)
@click.option("--name_split_on", help="Character to split path component on")
@click.option("--name_format", help="Str.format string for split output")
@click.option("--rank", "-r", help="Phylogenetic rank of matches")
@click.option("--parent_rank", "-pr", help="Phylogenetic rank of parent")
@click.option(
    "--parent_only",
    "-po",
    is_flag=True,
    default=False,
    help="Do not write low-level info",
)
@click.option(
    "--write_parent",
    "-wp",
    is_flag=True,
    default=True,
    help="Write top-level info",
)
@click.option(
    "--download",
    "-dl",
    is_flag=True,
    default=False,
    help="Download/decompress files",
)
@click.option(
    "--nucleic",
    "-nt",
    is_flag=True,
    default=False,
    help="Include nucleic-acid files",
)
@click.argument("uri")
@click.argument("parent_name")
@click.argument("outfile", type=click.Path(), nargs=-1)
def find_files(
    uri,  # file path, http URL, or site://
    parent_name,  # simple string, must not include spaces
    outfile,  # file path, optional, print if not used
    gff_pattern,  # globbing-style pattern, may be repeated
    faa_pattern,  # "
    fna_pattern,  # "
    exclude,  # "
    preference,  # ", order determines preference
    name,  # if specified, must be 1:1 with matches
    name_from_part,
    name_split_on,
    name_format,
    rank,  # applies to all
    parent_rank,
    write_parent,
    parent_only,
    download,
    nucleic,
):
    """
    Find files at URI and create input TOML file.

    \b
    Usage:
        azulejo find-files URI PARENT_NAME [OUTFILE]
        where:
            URI         ".", http URL, or recognized site:// tags
            PARENT_CODE Code for parent, without spaces
            OUTFILE     Name of TOML output file.  If not present,
                        output will be printed for debugging.
                        IF this file exists, entries in file will be
                        added to the exclude list.
    Example:
        azulejo find-files site://legfed/Glycine_max/ glyma glycine_max.toml

    """
    undeco_find_files(
        uri,
        parent_name,
        outfile,
        gff_pattern=gff_pattern,
        faa_pattern=faa_pattern,
        fna_pattern=fna_pattern,
        exclude=exclude,
        preference=preference,
        name=name,
        name_from_part=name_from_part,
        name_split_on=name_split_on,
        name_format=name_format,
        rank=rank,
        parent_rank=parent_rank,
        write_parent=write_parent,
        parent_only=parent_only,
        download=download,
        nucleic=nucleic,
    )


@cli.command()
@click_loguru.init_logger()
@click_loguru.log_elapsed_time()
@click.argument("input_toml")
def ingest(input_toml):
    """
    Marshal protein and genome sequence information.

    IDs must correspond between GFF and FASTA files and must be unique across
    the entire set.

    \b
    Example:
        azulejo ingest glyma+glyso.toml

    """
    undeco_ingest_sequences(input_toml, click_loguru=click_loguru)


@cli.command()
@click_loguru.init_logger()
@click.argument("setname")
@click.argument("synteny_type")
@click.argument("prefs", nargs=-1)
def proxy_genes(setname, synteny_type, prefs):
    """Calculate a set of proxy genes."""
    undeco_calculate_proxy_genes(setname, synteny_type, prefs)


@cli.command()
def bindir():
    """Print path to installed binaries."""
    print(f"{INSTALL_PATH}/bin")


@cli.command()
@click_loguru.init_logger(logfile=False)
@click.argument("dagchainer_tool_args", nargs=-1)
def dagchainer_tool(dagchainer_tool_args):
    """Run dagchainer_tool script."""
    dagchainer_tool = sh.Command(
        "dagchainer-tool.sh", search_paths=SEARCH_PATHS
    )
    try:
        dagchainer_tool(dagchainer_tool_args, _out=sys.stdout, _err=sys.stderr)
    except sh.ErrorReturnCode_1:
        logger.error(
            f"\"dagchainer-tool.sh {' '.join(dagchainer_tool_args)}\" failed"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# stdlib imports
import sys
from pathlib import Path

# third-party imports
import click
import pandas as pd


@click.command()
@click.argument("synfile", type=click.Path(readable=True))
@click.argument("homfile", type=click.Path(readable=True))
def join_homology_synteny(synfile, homfile):
    "From a 2-column tsv INFILE, output a frequency-sorted file"
    synpath = Path(synfile)
    outpath = synpath.parent / (
        synpath.name[: -len(synpath.suffix)] + ".hom.tsv"
    )
    syn = pd.read_csv(synpath, sep="\t")
    hom = pd.read_csv(homfile, sep="\t")
    hom = hom.set_index("members")
    syn["hom_cluster"] = syn["members"].map(hom["cluster_id"])
    syn.to_csv(outpath, sep="\t")
    n_clusts = 0
    n_consistent = 0
    for cluster_id, subframe in syn.groupby(by=["cluster_id"]):
        n_clusts += 1
        hom_counts = subframe["cluster_id"].value_counts()
        if len(hom_counts) == 1:
            n_consistent += 1
    n_inconsistent = n_clusts - n_consistent
    print(f"   Inconsistent:\t{n_inconsistent}")


if __name__ == "__main__":
    join_homology_synteny()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# stdlib imports
from pathlib import Path

# third-party imports
import click
import networkx as nx
import numpy as np
import pandas as pd


@click.command()
@click.argument("infile", type=click.Path(readable=True))
@click.argument("outfile", type=click.Path(writable=True))
def pairs_to_adjacency(infile, outfile):
    "Using a 2-column tsv of edges, output a file of adjacencies"
    outpath = Path(outfile)
    summarypath = outpath.parent / (
        outpath.name[: -len(outpath.suffix)] + "_summary.tsv"
    )
    histpath = outpath.parent / (
        outpath.name[: -len(outpath.suffix)] + "_hist.tsv"
    )
    G = nx.read_edgelist(infile)
    c = sorted(nx.connected_components(G), key=len, reverse=True)
    fh = outpath.open("w")
    fh.write("idx\tcluster_id\tsize\tmembers\n")
    n_items = 0
    count_list = []
    id_list = []
    for i, comp in enumerate(c):
        component = np.sort(pd.Index(list(comp)).to_numpy())
        id_list.append(i)
        size = len(comp)
        count_list.append(size)
        for node in component:
            fh.write(f"{n_items}\t{i}\t{size}\t{node}\n")
            n_items += 1
    fh.close()
    print(f"   Items:\t{n_items}")
    n_clusts = len(count_list)
    print(f"   Clusters:\t{n_clusts}")
    del G, c
    cluster_counts = pd.DataFrame({"size": count_list})
    largest_cluster = cluster_counts["size"].max()
    print(f"   Largest:\t{largest_cluster}")
    cluster_hist = (
        pd.DataFrame(cluster_counts.value_counts()).sort_index().reset_index()
    )
    cluster_hist = cluster_hist.set_index("size")
    cluster_hist = cluster_hist.rename(columns={0: "n"})
    cluster_hist["item_pct"] = (
        cluster_hist["n"] * cluster_hist.index * 100.0 / n_items
    )
    cluster_hist.to_csv(histpath, sep="\t", float_format="%5.2f")
    cluster_hist["cluster_pct"] = cluster_hist["n"] * 100.0 / n_clusts
    cluster_hist.to_csv(histpath, sep="\t", float_format="%5.2f")
    clusters = pd.DataFrame({"anchor.id": id_list, "count": count_list})
    clusters.to_csv(summarypath, sep="\t")


if __name__ == "__main__":
    pairs_to_adjacency()



# -*- coding: utf-8 -*-
"""External merges for synteny operations."""

# standard library imports
import array

# from os.path import commonprefix as prefix
from pathlib import Path

# third-party imports
import numpy as np
import pandas as pd

# helper_functions
def _unpack_payloads(vec):
    """Unpack TSV ints in payload."""
    values = np.array(
        [[int(i) for i in s.split("\t")] for s in vec.compressed()]
    ).transpose()
    return values


class AmbiguousMerger(object):
    """Counts instance of merges."""

    def __init__(
        self,
        count_key="value",
        ordinal_key="count",
        ambig_key="ambig",
        ambig_count_key=None,
        start_base=0,
        ambig_ordinal_key="count.ambig",
        alt_hash=False,
    ):
        """Create arrays as instance attributes."""
        self.count_key = count_key
        self.ordinal_key = ordinal_key
        self.start_base = start_base
        self.ambig_key = ambig_key
        self.ambig_ordinal_key = ambig_ordinal_key
        self.values = array.array("L")
        self.counts = array.array("h")
        self.ambig = array.array("h")
        self.alt_hash = alt_hash
        self.ambig_count_key = ambig_count_key
        if self.alt_hash:
            self.alt_hash_dict = {}
            self.count_dict = {}
            self.ambig_dict = {}
            self.id_no = self.start_base

    def merge_func(self, value, count, payload_vec):
        """Return list of merged values."""
        self.values.append(value)
        self.counts.append(count)
        arr = _unpack_payloads(payload_vec)
        max_ambig = arr[0].max()
        self.ambig.append(max_ambig)
        if self.alt_hash:
            # if this is true, vec[1] contains a list
            # of alt hashes, with 0 as a fill value to be ignoreed
            self.alt_hash_dict[value] = list({i for i in arr[1] if i != 0})
            self.count_dict[value] = count
            self.ambig_dict[value] = max_ambig

    def results(self):
        """Calculate list of merges."""
        drop_list = []
        if self.alt_hash:
            for hash_val in self.alt_hash_dict:
                related_hashes = [hash_val] + [
                    alt
                    for alt in self.alt_hash_dict[hash_val]
                    if alt in self.count_dict
                ]
                if len(related_hashes) == 1:
                    continue
                related_hashes.sort()  # take the first in numberical order if all else is equal
                non_ambig_hashes = [
                    h for h in related_hashes if self.ambig_dict[h] == 1
                ]
                max_count_idx = np.argmax(
                    [self.count_dict[h] for h in non_ambig_hashes]
                )
                best_hash = non_ambig_hashes[max_count_idx]
                if best_hash != hash_val:
                    drop_list.append(hash_val)
            del self.alt_hash_dict, self.count_dict, self.ambig_dict
        merge_frame = pd.DataFrame(
            {self.count_key: self.counts, self.ambig_key: self.ambig},
            index=self.values,
            dtype=pd.UInt32Dtype(),
        )
        merge_frame.drop(drop_list, inplace=True)
        merge_frame.sort_values(
            by=[self.ambig_key, self.count_key], inplace=True
        )
        unambig_frame = merge_frame[merge_frame[self.ambig_key] == 1].copy()
        n_unambig = len(unambig_frame)
        unambig_frame[self.ordinal_key] = pd.array(
            range(self.start_base, self.start_base + n_unambig),
            dtype=pd.UInt32Dtype(),
        )
        del unambig_frame[self.ambig_key]
        ambig_frame = merge_frame[merge_frame[self.ambig_key] > 1].copy()
        del merge_frame
        if self.ambig_count_key is None:
            # Don't pass counts along
            del ambig_frame[self.count_key]
        else:
            ambig_frame = ambig_frame.rename(
                columns={self.count_key: self.ambig_count_key}
            )
        del ambig_frame[self.ambig_key]
        ambig_frame[self.ambig_ordinal_key] = pd.array(
            range(
                self.start_base + n_unambig,
                self.start_base + len(ambig_frame) + n_unambig,
            ),
            dtype=pd.UInt32Dtype(),
        )
        return unambig_frame, ambig_frame



# -*- coding: utf-8 -*-
"""Send and receive on-disk messages through names pips with file locking."""
# standard library imports
import contextlib
import fcntl
import sys
from pathlib import Path

# third-party imports
import attr
import numpy as np
import numpy.ma as ma
from memory_tempfile import MemoryTempfile

# module imports
from .common import SCRATCH_DEV
from .common import append_slash
from .common import disk_usage_mb
from .common import free_mb
from .common import is_writable

# global constants
MAX_LINE_LEN = 144


# shared functions


class SpaceAwareTempDevices:
    """Allocate and collect stats on the fastest device that will hold your data."""

    class TempDirAllocator:
        """Reserve space and collect usage stats on a space-aware temp device."""

        def __init(self, space_needed_mb=None):
            """Select the fastest device with enough memory."""
            self.space_needed = space_needed_mb
            self.dev = fallback_dir
            if self.space_needed is not None:
                if space_needed_mb < FREE_MEM_MB:
                    memdev_list = MemoryTempfile(
                        filesystem_types=["tmpfs", "shm"]
                    ).get_usable_mem_tempdir_paths()
                    for memdev in memdev_list:
                        self.initial_space_free = free_mb(memdev)
                        if self.initial_space_free >= space_needed_mb:
                            self.dev = memdev
                            break
                if self.dev == fallback_dir and SCRATCH_DEV is not None:
                    self.initial_space_free = free_mb(SCRATCH_DEV)
                    if self.initial_space_free >= space_needed_mb:
                        self.dev = SCRATCH_DEV
            if self.dev == fallback_dir:
                self.initial_space_free = free_mb(self.dev)

    def __init__(
        self, scratch_dev=None, fallback_dev="/tmp/", addl_fs_types=None
    ):
        """Create a list of writable tmp devices and space allocations on them."""
        self.fallback_dev = fallback_dev
        self.scratch_dev = scratch_dev
        fs_types = ["tmpfs", "shm"]
        if addl_fs_types is not None:
            fs_types += addl_fs_types
        self.memdev_list = [
            append_slash(dev)
            for dev in MemoryTempfile(
                filesystem_types=fs_types
            ).get_usable_mem_tempdir_paths()
            if is_writable(dev)
        ]
        logger.debug(f"Memory devices are {self.memdev_list}")
        if self.scratch_dev is not None:
            if not is_writable(self.scratch_dev):
                logger.warn(
                    f"Scratch device {self.scratch_dev} is not writable."
                )
                self.scratch_dev = None
            else:
                logger.debug(f"Scratch device is {self.scratch_dev}")
        if not is_writable(self.fallback_dev):
            logger.error(
                f"fallback working device {self.fallback_dev} is not writable!"
            )
            sys.exit(1)
        self.allocatable_devices = self.memdev_list.copy()
        if self.scratch_dev is not None:
            self.allocatable_devices.append(self.scratch_dev)
        self.allocations = {k: 0 for k in self.allocatable_devices}
        self.min_free = {k: free_mb(k) for k in self.allocatable_devices}


@attr.s
class DataMailboxes:
    """Pass data to and from on-disk FIFOs."""

    n_boxes = attr.ib()
    mb_dir_path = attr.ib(default=Path("./mailboxes/"))
    file_extension = attr.ib(default=None)

    def write_headers(self, header):
        """Initialize the mailboxes, writing a free-form header."""
        self.mb_dir_path.mkdir(parents=True, exist_ok=True)
        for i in range(self.n_boxes):
            mb_path = self.path_to_mailbox(i)
            with mb_path.open("w") as fh:
                fh.write(header)

    def write_tsv_headers(self, columns, index_name=None):
        """Initialize the mailboxes, writing a tab-separated header."""
        if index_name is None:
            start = "\t"
        else:
            start = f"{index_name}\t"
        colstring = "\t".join(columns)
        self.mb_dir_path.mkdir(parents=True, exist_ok=True)
        for i in range(self.n_boxes):
            mb_path = self.path_to_mailbox(i)
            with mb_path.open("w") as fh:
                fh.write(f"{start}{colstring}\n")

    @contextlib.contextmanager
    def locked_open_for_write(self, box_no):
        """Acquire a lock on a m."""
        mb_path = self.path_to_mailbox(box_no)
        with mb_path.open("a+") as fd:
            fcntl.flock(fd, fcntl.LOCK_EX)
            yield fd
            fcntl.flock(fd, fcntl.LOCK_UN)

    @contextlib.contextmanager
    def open_then_delete(self, box_no, delete=True):
        """Open a mailbox, then delete when done."""
        box_path = self.path_to_mailbox(box_no)
        with box_path.open("r") as fd:
            yield fd
            if delete:
                box_path.unlink()

    def path_to_mailbox(self, box_no):
        """Return a path to a mailbox file."""
        if self.file_extension is None:
            ext = ""
        else:
            ext = f".{self.file_extension}"
        return self.mb_dir_path / f"{box_no}{ext}"

    def delete(self):
        """Remove the mailbox directory."""
        file_list = list(self.mb_dir_path.glob("*"))
        for file in file_list:
            file.unlink()
        self.mb_dir_path.rmdir()


@attr.s
class ExternalMerge(object):
    """Merges integers from files."""

    file_path_func = attr.ib(default=None)
    n_merge = attr.ib(default=None)
    value_vec = None
    fh_list = []

    def init(self, header_value):
        """Read and check the header info."""
        self.fh_list = [
            self.file_path_func(i).open() for i in range(self.n_merge)
        ]
        headers = [next(fh).rstrip() for fh in self.fh_list]
        if headers != ([header_value] * self.n_merge):
            print("Error in header values.")
            sys.exit(1)
        self.value_vec = ma.masked_array(
            np.zeros(self.n_merge),
            mask=np.zeros(self.n_merge),
        ).astype(np.uint32)
        self.payloads = np.array(
            [""] * self.n_merge, dtype=f"<U{MAX_LINE_LEN}"
        )
        self._next_vals(np.arange(self.n_merge))

    def _next_vals(self, index_vec):
        """Return the value of the next iterated value."""
        for index in index_vec:
            try:
                splitline = next(self.fh_list[index]).split("\t")
                self.value_vec[index] = int(splitline[0])
                payload = "\t".join(splitline[1:]).strip()
                if len(payload) > MAX_LINE_LEN:
                    print(f"payload line too long in file index{index}")
                    sys.exit(1)
                self.payloads[index] = payload
            except StopIteration:
                self.value_vec.mask[index] = 1

    def _close_all(self):
        """Close all filehandles."""
        for i in range(self.n_merge):
            self.fh_list[i].close()

    def merge(self, merge_obj):
        """Call merge_obj.merge_func for each merge, return merge_obj.results()."""
        while (~self.value_vec.mask).sum() > 1:
            minimum = np.amin(self.value_vec)
            min_vec = self.value_vec == minimum
            where_min = np.where(min_vec)[0]
            count = len(where_min)
            if count > 1:
                merge_obj.merge_func(
                    minimum,
                    count,
                    ma.masked_array(self.payloads, mask=~min_vec),
                )
            self._next_vals(where_min)
        self._close_all()
        return merge_obj.results()



# -*- coding: utf-8 -*-
"""Install binary dependencies."""
# standard library imports
import os
import pkg_resources
import pkgutil
import platform
import shutil
import sys
import tempfile
from pathlib import Path
from packaging import version

# third-party imports
import sh
from progressbar import DataTransferBar
from requests_download import ProgressTracker
from requests_download import download as request_download

# module imports
from .common import MinSpaceTracker
from .common import BUILD_DEV
from .common import SEARCH_PATHS
from .common import free_mb
from .common import is_writable
from .common import logger

# Global constants
PLATFORM_LIST = ["linux", "bsd", "macos", "unknown"]
EXECUTABLE_EXTS = (".sh", ".pl", ".awk", ".py")
MEGABYTES = 1024.0 * 1024.0


def default_version_splitter(instring):
    """Split the version string out of version output."""
    return instring.split()[-1]


def walk_package(pkgname, root):
    """Walk through a package_resource."""
    dirs = []
    files = []
    for name in pkg_resources.resource_listdir(pkgname, str(root)):
        fullname = root / name
        if pkg_resources.resource_isdir(pkgname, str(fullname)):
            dirs.append(fullname)
        else:
            files.append(Path(name))
    for new_path in dirs:
        yield from walk_package(pkgname, new_path)
    yield root, dirs, files


class DependencyInstaller(object):
    """Install and check binary dependencies."""

    def __init__(
        self,
        dependency_dict,
        pkg_name=__name__,
        install_path=None,
        bin_path=None,
        force=False,
        accept_licenses=False,
        quiet=False,
    ):
        """Initialize dictionary of dependencies."""
        self.dependency_dict = dependency_dict
        self.force = force
        self.pkg_name = pkg_name
        self.dependencies = tuple(dependency_dict.keys())
        for dep in self.dependencies:
            if "required" not in self.dependency_dict[dep]:
                self.dependency_dict[dep]["required"] = True
        self.versions_checked = False
        if install_path is None:
            self.install_path = Path(sys.executable).parent.parent
        else:
            self.install_path = install_path
        if bin_path is None:
            self.bin_path = self.install_path / "bin"
        else:
            self.bin_path = bin_path
        self.bin_path_exists = self.bin_path.exists()
        self.install_path_writable = os.access(self.install_path, os.W_OK)
        system = platform.system()
        if system == "Linux":
            self.platform = "linux"
        elif system.endswith("BSD"):
            self.platform = "bsd"
        elif system == "Darwin":
            self.platform = "macos"
        else:
            self.platform = "unknown"
            logger.warning(
                f"Unknown platform {system}, installs will likely fail"
            )
        self.not_my_platform = PLATFORM_LIST.copy()
        self.not_my_platform.remove(self.platform)
        self.accept_licenses = accept_licenses
        self.quiet = quiet
        if quiet:
            self.output_kwargs = {}
        else:
            self.output_kwargs = {"_out": sys.stdout, "_err": sys.stderr}
        self.status_msg = None
        self.work_path = None
        self.build_path = None
        self.makeopts = []

    def check_all(self, exe_paths=False):
        """Check all depenencies for existence and version."""
        self.status_msg = ""
        for dep in self.dependencies:
            target_version = version.parse(
                self.dependency_dict[dep]["version"]
            )
            version_command = self.dependency_dict[dep]["version_command"]
            self.dependency_dict[dep]["installed"] = not self.force
            for binary in self.dependency_dict[dep]["binaries"]:
                if sh.which(binary, paths=SEARCH_PATHS) == None:
                    self.dependency_dict[dep]["installed"] = False
                    exe = binary
                    ver_str = "not installed."
                else:
                    exe = sh.Command(binary, search_paths=SEARCH_PATHS)
                    try:
                        ver_out = exe(
                            *version_command, _err_to_out=True
                        ).rstrip("\n")
                        if "version_parser" in self.dependency_dict[dep]:
                            installed_version = version.parse(
                                self.dependency_dict[dep]["version_parser"](
                                    ver_out
                                )
                            )
                        else:
                            installed_version = version.parse(
                                default_version_splitter(ver_out)
                            )
                    except sh.ErrorReturnCode_1:
                        installed_version = version.parse("0.0")
                    if installed_version == target_version:
                        ver_str = (
                            f"at recommended version {installed_version}."
                        )
                    elif installed_version < target_version:
                        ver_str = (
                            f"installed {installed_version} <  target"
                            f" {target_version}."
                        )
                        self.dependency_dict[dep]["installed"] = False
                    elif installed_version > target_version:
                        ver_str = (
                            f"installed {installed_version} exceeds target"
                            f" {target_version}."
                        )
                if exe_paths:
                    self.status_msg += f"{exe} {ver_str}\n"
                else:
                    self.status_msg += f"{binary} {ver_str}\n"
        self.versions_checked = True
        # Check that bin directory exists and is writable.
        if self.bin_path_exists:
            bin_path_state = "exists, "
        else:
            bin_path_state = "doesn't exist, "
        if self.install_path_writable:
            bin_path_state += "writable, "
        else:
            bin_path_state += "not writable, "
        all_required = all(
            [
                self.dependency_dict[d]["installed"]
                for d in self.dependency_dict
                if self.dependency_dict[d]["required"]
            ]
        )
        all_optional = all(
            [
                self.dependency_dict[d]["installed"]
                for d in self.dependency_dict
                if not self.dependency_dict[d]["required"]
            ]
        )
        if all_required:
            if all_optional:
                optional_str = "and optional"
            else:
                optional_str = ""
            self.status_msg += (
                f"All required {optional_str} dependencies are installed.\n"
            )
        return all_required

    def status(self, exe_paths=False):
        """Returns and the installation status message."""
        if self.status_msg is None:
            self.check_all(exe_paths=exe_paths)
        return self.status_msg

    def install_list(self, deplist):
        """Install needed dependencies from a list."""
        for dep in deplist:
            alldeps = list(self.dependency_dict.keys()) + ["all"]
            if dep not in alldeps:
                logger.error(f'"{dep}" is not a recognized dependency')
                logger.error(f"possible dependencies are {alldeps}")
                sys.exit(1)
        self.check_all()
        if deplist == ("all",):
            deplist = [
                d
                for d in self.dependencies
                if self.dependency_dict[d]["required"]
            ]
        install_list = [
            dep
            for dep in deplist
            if not self.dependency_dict[dep]["installed"]
        ]
        if len(install_list):
            if not self.bin_path_exists:
                logger.error(
                    f"Installation directory {self.install_path} does not"
                    " exist."
                )
                sys.exit(1)
            if not self.install_path_writable:
                logger.error(
                    f"Installation directory {self.install_path} is not"
                    " writable."
                )
                sys.exit(1)
        for dep in install_list:
            self.install(dep)

    def _get_makeopts(self):
        """Get options for make from MAKEOPTS envvar."""
        if "MAKEOPTS" in os.environ:
            self.makeopts = os.environ["MAKEOPTS"].split()

    def _git(self, dep):
        """Git clone from list."""
        git = sh.Command("git")
        for repo in self.dependency_dict[dep]["git_list"]:
            logger.debug(f"   cloning {dep} repo {repo}")
            git.clone(repo, **self.output_kwargs)

    def _download_untar(self, dep):
        """Download and untar tarball."""
        tar = sh.Command("tar")
        download_url = self.dependency_dict[dep]["tarball"]
        dlname = download_url.split("/")[-1]
        download_path = Path(".") / dlname
        logger.debug(f"downloading {dep} at {download_url} to {dlname}")
        if self.quiet:
            trackers = ()
        else:
            trackers = (ProgressTracker(DataTransferBar()),)
        request_download(download_url, download_path, trackers=trackers)
        logger.debug(
            f"downloaded file {download_path}, size"
            f" {download_path.stat().st_size}"
        )
        try:
            tar("xvf", download_path, **self.output_kwargs)
        except:
            logger.error(f"untar of {download_path} failed")
            sys.exit(1)

    def _download_binaries(self, dep):
        """Download and untar tarball."""
        download_dict = self.dependency_dict[dep]["download_binaries"]
        if self.platform in download_dict:
            download_url = download_dict[self.platform]
        else:
            logger.warning(
                f"No binaries for download for {dep}, will fake it."
            )
            return
        dlname = download_url.split("/")[-1]
        download_path = Path(".") / dlname
        logger.debug(f"downloading {dep} at {download_url} to {dlname}")
        if self.quiet:
            trackers = ()
        else:
            trackers = (ProgressTracker(DataTransferBar()),)
        request_download(download_url, download_path, trackers=trackers)
        logger.debug(
            f"downloaded file {download_path}, size"
            f" {download_path.stat().st_size}"
        )

    def _configure(self, dep):
        """Run configure to generate a makefile."""
        logger.debug(f"   configuring {dep} in {Path.cwd()}")
        config_args = [f"--prefix={self.install_path}"] + self.dependency_dict[
            dep
        ]["configure"]
        configure = sh.Command("./configure")
        logger.debug(f"   configure {config_args} in {Path.cwd()}")
        try:
            configure(config_args, **self.output_kwargs)
        except:
            logger.error(f"configure of {dep} failed.")
            sys.exit(1)

    def _cmake(self, dep):
        """Run cmake to generate makefiles."""
        self.build_path = Path(self.work_path.parent / "build")
        self.build_path.mkdir()
        cmake_args = [
            f"-DCMAKE_INSTALL_PREFIX={self.install_path}",
            "-S",
            ".",
            "-B",
            str(self.build_path),
        ] + self.dependency_dict[dep]["cmake"]
        logger.debug(f"   running cmake {cmake_args} on {dep} in {Path.cwd()}")
        cmake = sh.Command("cmake")
        try:
            cmake(cmake_args, **self.output_kwargs)
        except:
            logger.error(f"cmake of {dep} failed.")
            sys.exit(1)

    def _copy_in_files(self, out_head, pkgname, dep, force=True):
        """Copy any files under installer_data to build directory."""
        resource_path = Path("installer_data") / dep
        platform_path = resource_path / self.platform
        not_my_platform_path_parts = [
            (resource_path / p).parts for p in self.not_my_platform
        ]
        if pkg_resources.resource_exists(pkgname, str(resource_path)):
            for root, unused_dirs, files in walk_package(
                pkgname, resource_path
            ):
                root_parts = root.parts
                if root_parts[:3] in not_my_platform_path_parts:
                    continue
                elif root_parts[:3] == platform_path.parts:
                    subdir_parts = root_parts[3:]
                else:
                    subdir_parts = root_parts[2:]
                if len(subdir_parts) > 0 and subdir_parts[-1] == "__pycache__":
                    continue
                out_path = out_head.joinpath(*subdir_parts)
                if not out_path.exists() and len(files) > 0:
                    logger.debug(f'Creating "{str(out_path)}" directory')
                    out_path.mkdir(0o755, parents=True)
                for filename in files:
                    logger.debug(f"Copying {filename}")
                    try:
                        data_string = pkgutil.get_data(
                            __name__, str(root / filename)
                        ).decode("UTF-8")
                    except UnicodeDecodeError:
                        logger.warning(
                            f"File {filename} contained an undecodable string, skipping"
                        )
                        continue
                    file_path = out_path / filename
                    if file_path.exists() and not force:
                        logger.error(
                            f"File {str(file_path)} already exists."
                            + "  Use --force to overwrite."
                        )
                        sys.exit(1)
                    elif file_path.exists() and force:
                        operation = "Overwriting"
                    else:
                        operation = "Creating"
                    logger.debug(f"{operation} {file_path}")
                    with file_path.open(mode="wt") as fh:
                        fh.write(data_string)
                    if filename.suffix in EXECUTABLE_EXTS:
                        file_path.chmod(0o755)

    def _patch(self, dep):
        """Apply a patch."""
        patch = sh.Command("patch")
        for patch_file in self.dependency_dict[dep]["patches"]:
            patch_args = ["-p1", "-i", patch_file]
            logger.debug(f"   patch {patch_args} in {Path.cwd()}")
            try:
                patch(patch_args, **self.output_kwargs)
            except:
                logger.error(f"patch {patch_file} of {dep} failed.")
                sys.exit(1)

    def _make(self, dep):
        """Run make to build package."""
        make = sh.Command("make")
        make_args = self.dependency_dict[dep]["make"] + self.makeopts
        logger.debug(f"   make {make_args} in {Path.cwd()}")
        try:
            make(make_args, **self.output_kwargs)
        except:
            logger.error(f"make of {dep} failed.")
            sys.exit(1)

    def _make_install(self, dep):
        """Run make install to install a package."""
        make = sh.Command("make")
        make_args = self.dependency_dict[dep]["make_install"] + self.makeopts
        logger.debug(f"   make install {make_args} in {Path.cwd()}")
        try:
            make.install(make_args, **self.output_kwargs)
        except:
            logger.error(f"make install of {dep} failed.")

    def _copy_binaries(self, dep):
        """Copy the named binary to the bin directory."""
        logger.info(f"   copying {dep} into {self.bin_path}")
        for binary in self.dependency_dict[dep]["copy_binaries"]:
            binpath = Path(binary)
            shutil.copy2(binpath, self.bin_path / binpath.name)

    def _check_for_license_acceptance(self, dep):
        """Prompt for acceptance of license terms."""
        if "license" in self.dependency_dict[dep]:
            license_name = self.dependency_dict[dep]["license"]
        else:
            license_name = "restrictive"
        if "license_file" in self.dependency_dict[dep]:
            license_text = Path(
                self.dependency_dict[dep]["license_file"]
            ).read_text()
            logger.warning(license_text)
        while "invalid answer":
            reply = (
                str(
                    input(
                        f"Do you accept this {license_name} license? (y/n): "
                    )
                )
                .lower()
                .strip()
            )
            if len(reply) > 0:
                if reply[0] == "y":
                    return True
                if reply[0] == "n":
                    return False

    def install(self, dep):
        """Install a particular dependency."""
        logger.info(f"installing {dep}")
        dep_dict = self.dependency_dict[dep]
        build_dev_path = BUILD_DEV
        if not is_writable(build_dev_path):
            logger.error(f"Build directory {build_dev_path} is not writable")
            sys.exit(1)
        initial_free_mb = free_mb(build_dev_path)
        free_tracker = MinSpaceTracker(build_dev_path)
        logger.debug(
            f"building on {build_dev_path} with {free_tracker.report_min()} MB free"
        )
        with tempfile.TemporaryDirectory(
            prefix="installer-", dir=build_dev_path
        ) as tmp:
            tmppath = Path(tmp)
            logger.debug(f'build directory: "{tmppath}"')
            os.chdir(tmppath)
            #
            # Get the sources.  Alternatives are git or download
            #
            if "git_list" in dep_dict:
                self._git(dep)
            elif "tarball" in dep_dict:
                self._download_untar(dep)
            elif "download_binaries" in dep_dict:
                self._download_binaries(dep)
            free_tracker.check()
            #
            # Change to the work directory.
            #
            logger.debug(f'building in directory {dep_dict["dir"]}')
            dirpath = Path(".") / dep_dict["dir"]
            if not dirpath.exists():
                raise ValueError(f'directory "{dirpath}" does not exist.')
            if not dirpath.is_dir():
                raise ValueError(f'directory "{dirpath}" is not a directory.')
            os.chdir(dirpath)
            self.work_path = Path.cwd()
            self.build_path = self.work_path
            #
            # Copy in any additional files.
            #
            self._copy_in_files(self.work_path, self.pkg_name, dep, force=True)
            #
            # Check licensing
            #
            if "license" in dep_dict:
                if (
                    "license_restrictive" in dep_dict
                    and dep_dict["license_restrictive"]
                ):
                    logger.warning(
                        f"{dep} has a restrictive {dep_dict['license']} license"
                    )
                    if not self.accept_licenses:
                        accept = self._check_for_license_acceptance(dep)
                        if not accept:
                            logger.error(
                                f"License terms for {dep} not accepted, skipping installation"
                            )
                            return
                else:
                    logger.info(f"{dep} has a {dep_dict['license']} license")
            #
            # Build the executables.
            #
            self._get_makeopts()
            if "patches" in dep_dict:
                self._patch(dep)
            if "cmake" in dep_dict:
                self._cmake(dep)
            if "configure" in dep_dict:
                os.chdir(self.build_path)
                self._configure(dep)
            if "configure_extra_dirs" in dep_dict:
                for newdir in dep_dict["configure_extra_dirs"]:
                    os.chdir(self.build_path / newdir)
                    self._configure(dep)
                    os.chdir(self.build_path)
            if "make" in dep_dict:
                os.chdir(self.build_path)
                self._make(dep)
            if "make_extra_dirs" in dep_dict:
                for newdir in dep_dict["make_extra_dirs"]:
                    os.chdir(self.build_path / newdir)
                    self._make(dep)
                    os.chdir(self.build_path)
            free_tracker.check()
            #
            # Install the executables.
            #
            if "make_install" in dep_dict:
                os.chdir(self.build_path)
                self._make_install(dep)
            elif "copy_binaries" in dep_dict:
                os.chdir(self.build_path)
                self._copy_binaries(dep)
            logger.info(
                f"Build of {dep} on {build_dev_path} used"
                + f" {free_tracker.report_used()}/{free_tracker.initial_space} MB"
            )



# -*- coding: utf-8 -*-
"""Sequence (FASTA) and genome (GFF) ingestion operations."""
# standard library imports
import contextlib
import json
import os
import shutil
import sys
import tempfile
from fnmatch import fnmatch
from pathlib import Path
from urllib.request import Request, urlopen, urlretrieve

# third-party imports
import attr
import dask.bag as db
import gffpandas.gffpandas as gffpd
import numpy as np
import pandas as pd
import toml
from bs4 import BeautifulSoup
from dask.diagnostics import ProgressBar

# first-party imports
import smart_open
from pathvalidate import ValidationError
from pathvalidate import validate_filename as pv_validate_filename

# module imports
from .common import ALTERNATE_ABBREV
from .common import CHROMOSOME_ABBREV
from .common import CHROMOSOME_SYNONYMS
from .common import DIRECTIONAL_CATEGORY
from .common import FRAGMENTS_FILE
from .common import PLASTID_STARTS
from .common import PROTEINS_FILE
from .common import PROTEOMES_FILE
from .common import SAVED_INPUT_FILE
from .common import SCAFFOLD_ABBREV
from .common import SCAFFOLD_SYNONYMS
from .common import bool_to_y_or_n
from .common import dotpath_to_path
from .common import logger
from .common import sort_proteome_frame
from .common import y_or_n_to_bool
from .common import write_tsv_or_parquet
from .core import cleanup_fasta
from .taxonomy import rankname_to_number

# global constants
__all__ = ["read_from_url", "ingest_sequences"]
FILE_TRANSPORT = "file://"
REQUIRED_LEAF_NAMES = (
    "fasta",
    "gff",
)
COMPRESSION_EXTENSIONS = (
    "gz",
    "bz2",
)
POSSIBLE_FEATURES = ("mRNA", "CDS")
POSSIBLE_ID_COLS = ("ID", "gene", "Name", "protein_id", "Parent")
MINIMUM_PROTEINS = 100

SITES = {
    "legfed": {
        "url": "https://v1.legumefederation.org/data/index/public/",
        "faa_pattern": ("*protein_primaryTranscript.faa.gz",),
        "fna_pattern": ("*transcript_primaryTranscript.fna.gz",),
        "gff_pattern": ("*gene_models_main.gff3.gz",),
        "name_from_part": 1,
        "name_split_on": ".",
        "name_format": "{0}",
    }
}

# helper functions
def _remove_leading_zeroes_in_field(string):
    """If blank-separated fields are integer, remove the leading zero."""
    split = string.split(" ")
    for i, field in enumerate(split):
        if field.isdigit():
            split[i] = f"{int(field)}"
    return " ".join(split)


def _is_scaffold(ids):
    """Assess whether fragment is likely a scaffold."""
    return bool_to_y_or_n(
        np.logical_or(
            ids.str.contains(SCAFFOLD_ABBREV),
            ids.str.startswith(ALTERNATE_ABBREV),
        )
    )


def _is_chromosome(ids):
    """Assess whether fragment is likely a chromosome."""
    return bool_to_y_or_n(
        np.logical_and(
            ids.str.startswith(CHROMOSOME_ABBREV),
            ~y_or_n_to_bool(_is_scaffold(ids)),
        )
    )


def _is_plastid(ids):
    """Assess whether fragment is likely a plastid."""
    return bool_to_y_or_n(
        np.logical_or.reduce([ids.str.startswith(s) for s in PLASTID_STARTS])
    )


def _validate_filename(name):
    """Check if a potential filename is valid or not."""
    try:
        pv_validate_filename(name)
    except ValidationError as e:
        logger.error(f"Invalid component name {name} in input file")
        logger.error(e)
        sys.exit(1)
    return name


def _strip_file_uri(url):
    """Removes the file:// uri from a URL string."""
    if url.startswith(FILE_TRANSPORT):
        return url[len(FILE_TRANSPORT) :]
    return url


def _validate_uri(uri):
    """Check if the transport at the start of a URI is valid or not."""
    try:
        smart_open.parse_uri(uri)
    except NotImplementedError:
        logger.error(f'Unimplemented transport in uri "{uri}"')
        sys.exit(1)
    return uri


def ingest_sequences(input_toml, click_loguru=None):
    """Marshal protein and genome sequence information."""
    options = click_loguru.get_global_options()
    user_options = click_loguru.get_user_global_options()
    parallel = user_options["parallel"]
    input_obj = TaxonomicInputTable(Path(input_toml), write_table=False)
    input_table = input_obj.input_table
    logger.info(f"Output directory: {input_obj.setname}/")
    set_path = Path(input_obj.setname)
    arg_list = []
    for unused_i, row in input_table.iterrows():
        arg_list.append(
            (
                row["path"],
                row["fasta_url"],
                row["gff_url"],
            )
        )
    bag = db.from_sequence(arg_list)
    file_stats = []
    if not options.quiet:
        logger.info(f"Extracting FASTA/GFF info for {len(arg_list)} genomes:")
        ProgressBar().register()
    if parallel:
        file_stats = bag.map(
            read_fasta_and_gff, verbose=options.verbose
        ).compute()
    else:
        for args in arg_list:
            file_stats.append(
                read_fasta_and_gff(args, verbose=options.verbose)
            )
    del arg_list
    seq_stats = pd.DataFrame.from_dict([s[0] for s in file_stats]).set_index(
        "path"
    )
    frag_stats = pd.DataFrame.from_dict([s[1] for s in file_stats]).set_index(
        "path"
    )
    proteomes = pd.concat(
        [input_table.set_index("path"), frag_stats, seq_stats], axis=1
    )
    proteomes.drop(["fasta_url", "gff_url"], axis=1, inplace=True)
    proteomes = sort_proteome_frame(proteomes)
    if not options.quiet:
        with pd.option_context(
            "display.max_rows", None, "display.float_format", "{:,.2f}%".format
        ):
            print(
                proteomes.drop(
                    [
                        col
                        for col in proteomes.columns
                        if col.startswith("phy")
                    ],
                    axis=1,
                )
            )
    proteome_table_path = set_path / PROTEOMES_FILE
    logger.info(
        f'Edit table of proteomes at "{proteome_table_path}"'
        + " to change preferences"
    )
    write_tsv_or_parquet(proteomes, proteome_table_path)
    idx_start = 0
    for df in [s[2] for s in file_stats]:
        df.index = range(idx_start, idx_start + len(df))
        idx_start += len(df)
    frags = pd.concat([s[2] for s in file_stats], axis=0)
    frags.index.name = "idx"
    fragalyzer = FragmentCharacterizer()
    frags = fragalyzer.assign_frag_properties(frags)
    frags_path = set_path / FRAGMENTS_FILE
    if not frags_path.exists():
        logger.info(
            f'Edit fragment table at "{frags_path}" to rename fragments'
        )
        write_tsv_or_parquet(frags, frags_path)
    else:
        new_frags_path = set_path / ("new." + FRAGMENTS_FILE)
        logger.info(f'A fragments file table already exists at "{frags_path}"')
        logger.info(f'A new file has been written at "{new_frags_path}".')
        logger.info("Edit and rename it to rename fragments.")
        write_tsv_or_parquet(frags, new_frags_path)


def read_fasta_and_gff(args, verbose=False):
    """Read corresponding sequence and position files and construct consolidated tables."""
    dotpath, fasta_url, gff_url = args
    out_path = dotpath_to_path(dotpath)
    with read_from_url(fasta_url) as fasta_fh:
        unused_stem, unused_path, fasta_props, proteome_stats = cleanup_fasta(
            out_path, fasta_fh, dotpath, write_fasta=False, write_stats=False
        )
    n_fasta = len(fasta_props)
    if n_fasta < MINIMUM_PROTEINS:
        logger.error(f"For FASTA file at URL'{fasta_url}")
        logger.error(
            f"Number of proteins read {len(fasta_props)} is too small"
        )
        sys.exit(1)
    with filepath_from_url(gff_url) as local_gff_file:
        annotation = gffpd.read_gff3(local_gff_file)
    # We prefer mRNAs to CDS for reasons related to overlaps,
    # but non-euks don't usually have mRNA features in GFF.
    # Sometimes features like CDS are near-duplicates, so take
    # the first entry.
    n_features = 0
    for feat_type in POSSIBLE_FEATURES:
        filtered_features = annotation.filter_feature_of_type([feat_type])
        n_features = len(filtered_features.df)
        if n_features > MINIMUM_PROTEINS:
            if verbose:
                logger.debug(
                    f"Using {n_features} features of type '{feat_type}' in {gff_url}"
                )
            break
    if n_features < MINIMUM_PROTEINS:
        logger.error(
            f"Not enough {n_features} of types"
            + f" {POSSIBLE_FEATURES} in GFF file {gff_url}"
        )
        sys.exit(1)
    proteome_stats["gff.feature"] = feat_type
    try:
        non_uniq_feat_cols = filtered_features.attributes_to_columns()
    except ValueError as val_err:
        logger.error(val_err)
        logger.error(
            f"Badly-formed {feat_type} features in GFF file {gff_url}"
        )
        sys.exit(1)
    # sometimes gene names are in ID, other times they're in gene
    n_joint = 0
    n_missing_in_gff = 0
    for id_col in POSSIBLE_ID_COLS:
        if id_col not in non_uniq_feat_cols.columns:
            continue
        feat_cols = non_uniq_feat_cols.drop_duplicates(subset=[id_col])
        in_both = feat_cols[id_col].isin(fasta_props.index)
        n_joint = in_both.astype(int).sum()
        if n_joint > MINIMUM_PROTEINS:
            if verbose:
                logger.debug(
                    f"Found {n_joint} FASTA IDs in GFF feature '{id_col}'"
                )
            n_missing_in_gff = n_fasta - n_joint
            if n_missing_in_gff > 0:
                logger.warning(
                    f"{n_missing_in_gff} FASTA IDs not found in GFF file {gff_url}"
                )
            break
    n_uniq_features = len(feat_cols)
    proteome_stats["gff.dups"] = n_features - n_uniq_features
    if n_joint < MINIMUM_PROTEINS:
        logger.error(
            f"Not enough ID's found ({n_joint}) that match FASTA IDS"
            + f" from {POSSIBLE_ID_COLS} features in GFF file {gff_url}"
        )
        sys.exit(1)
    proteome_stats["gff.id"] = id_col
    proteome_stats["gff.missing"] = n_missing_in_gff
    # Drop any features not found in sequence file, e.g., zero-length
    features = feat_cols[in_both]
    del annotation, filtered_features, feat_cols, in_both, non_uniq_feat_cols
    features.drop(
        features.columns.drop(
            ["seq_id", "start", "strand", id_col]
        ),  # drop EXCEPT these
        axis=1,
        inplace=True,
    )  # drop non-essential columns
    features = features.set_index(id_col)
    features = features.rename(
        columns={
            "start": "frag.start",
            "seq_id": "tmp.seq_id",
            "strand": "tmp.strand",
        }
    )
    features.index.name = "prot.id"
    # Make categoricals
    features["frag.id"] = pd.Categorical(features["tmp.seq_id"])
    features["frag.direction"] = pd.Categorical(
        features["tmp.strand"], dtype=DIRECTIONAL_CATEGORY
    )
    # Drop any features not found in sequence file, e.g., zero-length
    features = features[features.index.isin(fasta_props.index)]
    # sort fragments by largest value
    frag_counts = features["frag.id"].value_counts()
    n_frags = len(frag_counts)
    frags = pd.DataFrame(
        {
            "path": [dotpath] * n_frags,
            "frag.len": frag_counts,
            "frag.orig_id": frag_counts.index,
        },
        index=frag_counts.index,
    )
    frag_stats = {
        "path": dotpath,
        "frag.n": n_frags,
        "frag.max": frag_counts[0],
    }
    features["frag.prot_count"] = pd.array(
        features["frag.id"].map(frag_counts)
    )
    features.sort_values(
        by=["frag.prot_count", "frag.id", "frag.start"],
        ascending=[False, False, True],
        inplace=True,
    )
    frag_id_range = []
    for frag_id in frag_counts.index:
        frag_id_range += list(range(frag_counts[frag_id]))
    features["frag.pos"] = frag_id_range
    del frag_id_range
    # join GFF info to FASTA info
    joined_path = out_path / PROTEINS_FILE
    features = features.join(fasta_props)
    write_tsv_or_parquet(features, joined_path, sort_cols=False)
    return proteome_stats, frag_stats, frags


@attr.s
class FragmentCharacterizer:
    """Rename and characterize fragments based on those names."""

    extra_plastid_starts = attr.ib(default=[])
    extra_chromosome_starts = attr.ib(default=[])
    extra_scaffold_starts = attr.ib(default=[])

    @attr.s
    class UnknownsAsScaffolds(object):
        """Keep track of unknown scaffolds."""

        scaf_dict = attr.ib(default={})
        n_scafs = attr.ib(default=0)
        extra_plastid_starts = attr.ib(default=[])

        def rename_unknowns(self, string):
            """If the begginning of the id isn't recognized, rename as unique scaffold."""
            for start in (
                [SCAFFOLD_ABBREV, CHROMOSOME_ABBREV]
                + PLASTID_STARTS
                + self.extra_plastid_starts
            ):
                if string.startswith(start):
                    return string
            if string not in self.scaf_dict:
                self.scaf_dict[string] = self.n_scafs
                self.n_scafs += 1
            return f"{ALTERNATE_ABBREV} {self.scaf_dict[string]}"

    @attr.s
    class PreserveUnique(object):
        """Ensure uniqueness of a series is preserved."""

        ser = attr.ib(default=None)

        def ensure_still_unique(self, new_ser):
            if new_ser.nunique() == self.ser.nunique():
                self.ser = new_ser

    def cleanup_frag_ids(self, orig_ids):
        """Try to automatically put genome fragments on a common basis."""
        uniq = self.PreserveUnique(ser=orig_ids)
        split_ids = orig_ids.str.split(".", expand=True)
        split_ids = split_ids.drop(
            [i for i in split_ids.columns if len(set(split_ids[i])) == 1],
            axis=1,
        )
        uniq.ensure_still_unique(split_ids.agg(".".join, axis=1))
        uniq.ensure_still_unique(uniq.ser.str.lower())
        uniq.ensure_still_unique(uniq.ser.str.lower().str.replace("_", " "))
        for synonym in CHROMOSOME_SYNONYMS:
            uniq.ensure_still_unique(
                uniq.ser.str.replace(synonym, CHROMOSOME_ABBREV)
            )
        uniq.ensure_still_unique(
            uniq.ser.str.replace(CHROMOSOME_ABBREV, CHROMOSOME_ABBREV + " ")
        )
        for synonym in SCAFFOLD_SYNONYMS:
            uniq.ensure_still_unique(
                uniq.ser.str.replace(synonym, SCAFFOLD_ABBREV)
            )
        uniq.ensure_still_unique(
            uniq.ser.str.replace(SCAFFOLD_ABBREV, SCAFFOLD_ABBREV + " ")
        )
        uniq.ensure_still_unique(uniq.ser.str.replace("  ", " "))
        clean = uniq.ser.map(_remove_leading_zeroes_in_field)
        sc_namer = self.UnknownsAsScaffolds()
        clean = clean.map(sc_namer.rename_unknowns)
        orig_ids.index = clean
        return clean

    def assign_frag_properties(self, frags):
        """Assign properties to fragments based on cleaned-up names."""
        clean_list = []
        for unused_path, subframe in frags.groupby(by=["path"]):
            clean_list.append(self.cleanup_frag_ids(subframe["frag.orig_id"]))
        clean_ids = pd.concat(clean_list).sort_index()
        frags["frag.is_plas"] = _is_plastid(clean_ids)
        frags["frag.is_scaf"] = _is_scaffold(clean_ids)
        frags["frag.is_chr"] = _is_chromosome(clean_ids)
        frags["frag.id"] = clean_ids
        return frags


class TaxonomicInputTable:
    """Parse an azulejo input dictionary."""

    def __init__(self, toml_path, write_table=True):
        """Create structures."""
        self.depth = 0
        try:
            tree = toml.load(toml_path)
        except TypeError:
            logger.error(f"Error in filename {toml_path}")
            sys.exit(1)
        except toml.TomlDecodeError as e:
            logger.error(f"File {toml_path} is not valid TOML")
            logger.error(e)
            sys.exit(1)
        if len(tree) > 1:
            logger.error(
                f"Input file {toml_path} should define a single "
                + f"object, but defines {len(tree)} instead"
            )
            sys.exit(1)
        self.setname = _validate_filename(list(tree.keys())[0])
        root_path = Path(self.setname)
        if not root_path.exists():
            logger.debug(f"Creating directory for set {self.setname}")
            root_path.mkdir(parents=True)
        self._Node = attr.make_class(
            "Node", ["path", "name", "rank", "rank_val"]
        )
        self._nodes = []
        self._genome_dir_dict = {}
        self._n_genomes = 0
        self._walk(self.setname, tree[self.setname])
        self.input_table = pd.DataFrame.from_dict(
            self._genome_dir_dict
        ).transpose()
        del self._genome_dir_dict
        del self._nodes
        self.input_table.index.name = "order"
        if write_table:
            input_table_path = root_path / PROTEOMES_FILE
            logger.debug(
                f"Input table of {len(self.input_table)} genomes written to"
                f" {input_table_path}"
            )
            write_tsv_or_parquet(self.input_table, input_table_path)
        saved_input_path = root_path / SAVED_INPUT_FILE
        if toml_path != saved_input_path:
            shutil.copy2(toml_path, root_path / SAVED_INPUT_FILE)

    def _walk(self, node_name, tree):
        """Recursively walk tree structure."""
        # Check for required field properties.
        if len(self._nodes) > 0:
            dot_path = f"{self._nodes[-1].path}.{node_name}"
        else:
            dot_path = node_name
        if "name" not in tree:
            tree["name"] = f"'{node_name}'"
        if "rank" not in tree:
            logger.error(f'Required entry "rank" not in entry {dot_path}')
            sys.exit(1)
        try:
            rank_val = rankname_to_number(tree["rank"])
        except ValueError as e:
            logger.error(f"Unrecognized taxonomic rank {tree['rank']}")
            logger.error(e)
            sys.exit(1)
        if (len(self._nodes) > 0) and rank_val <= self._nodes[-1].rank_val:
            logger.error(
                f"rank {tree['rank']} value {rank_val} is not less than"
                + f" previous rank value of {self._nodes[-1].rank_val}"
            )
            sys.exit(1)
        # Push node onto stack
        self._nodes.append(
            self._Node(
                _validate_filename(dot_path),
                tree["name"],
                tree["rank"],
                rank_val,
            )
        )
        self.depth = max(self.depth, len(self._nodes))
        # Initialize node properties dictionary
        properties = {"path": dot_path, "children": []}
        for k, v in tree.items():
            if isinstance(v, dict):
                properties["children"] += [k]
                self._walk(k, v)
            else:
                properties[k] = v
        if len(properties["children"]) == 0:
            del properties["children"]
        # Check if this is a genome directory node
        genome_dir_properties = [
            (p in properties) for p in REQUIRED_LEAF_NAMES
        ]
        if any(genome_dir_properties):
            if not all(genome_dir_properties):
                missing_properties = [
                    p
                    for i, p in enumerate(REQUIRED_LEAF_NAMES)
                    if not genome_dir_properties[i]
                ]
                logger.error(
                    f"Missing properties {missing_properties} "
                    + f"for node {dot_path}"
                )
                sys.exit(1)
            if "uri" not in tree:
                uri = FILE_TRANSPORT
            else:
                uri = _validate_uri(tree["uri"])
                if not uri.endswith("/"):
                    uri += "/"
            self._genome_dir_dict[self._n_genomes] = {"path": dot_path}
            if "preference" not in tree:
                self._genome_dir_dict[self._n_genomes]["preference"] = ""
            else:
                self._genome_dir_dict[self._n_genomes]["preference"] = tree[
                    "preference"
                ]
            for n in self._nodes:
                self._genome_dir_dict[self._n_genomes][
                    f"phy.{n.rank}"
                ] = n.name
            self._genome_dir_dict[self._n_genomes][
                "fasta_url"
            ] = _strip_file_uri(uri + tree["fasta"])
            self._genome_dir_dict[self._n_genomes][
                "gff_url"
            ] = _strip_file_uri(uri + tree["gff"])
            self._n_genomes += 1
        for n in self._nodes:
            properties[n.rank] = n.name
        node_path = dotpath_to_path(dot_path)
        node_path.mkdir(parents=True, exist_ok=True)
        properties_file = node_path / "node_properties.json"
        logger.debug(f"Writing properties file to {properties_file}")
        with properties_file.open("w") as filepointer:
            json.dump(properties, filepointer)
        # Pop node from stack
        self._nodes.pop()


@contextlib.contextmanager
def _cd(newdir, cleanup=lambda: True):
    """Change directory with cleanup."""
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)
        cleanup()


@contextlib.contextmanager
def read_from_url(url):
    """Read from a URL transparently decompressing if compressed."""
    yield smart_open.open(url)


@contextlib.contextmanager
def filepath_from_url(url):
    """Get a local file from a URL, decompressing if needed."""
    filename = url.split("/")[-1]
    compressed = False
    uncompressed_filename = filename
    for ext in COMPRESSION_EXTENSIONS:
        if filename.endswith(ext):
            compressed = True
            uncompressed_filename = filename[: -(len(ext) + 1)]
            break
    if (
        url.find("://") == -1 and not compressed
    ):  # no transport, must be a file
        yield url
    else:
        dirpath = tempfile.mkdtemp()
        filehandle = smart_open.open(url)
        dldata = filehandle.read()

        def cleanup():
            shutil.rmtree(dirpath)

        with _cd(dirpath, cleanup):

            with open(uncompressed_filename, "w") as f:
                f.write(dldata)
            tmpfile = str(Path(dirpath) / uncompressed_filename)
            yield tmpfile


def _path_to_name(path_str, name_from_part, name_split_on, name_format):
    """Extract a genome name from a URI path."""
    path = Path(path_str)
    component_no = -1 - int(name_from_part)
    component = path.parts[component_no]
    if name_split_on == "":
        split = [component]
    else:
        split = component.split(name_split_on)
    return name_format.format(*split)


def _replace_spaces(url):
    """Replace spaces in a string with HTML escape."""
    return url.replace(" ", "%20")


def _is_dir(path):
    """Return true if path is a directory."""
    if path[-1] == "/" and path[0] != ".":
        return True
    return False


def _url_paths(url, base_len=None):
    """Read paths recursively from an http URL."""
    paths = []
    url = _replace_spaces(url)
    if base_len is None:
        base_len = len(url)
    try:
        resp = urlopen(Request(url)).read()
    except:
        logger.error(f"Unable to retrieve URL {url}")
        sys.exit(1)
    try:
        soup = BeautifulSoup(resp, "html.parser")
    except:
        logger.error("Unable to parse response from {uri}")
        sys.exit(1)
    for anchor in soup.find_all("a"):
        new_path = anchor.extract().get_text()
        if _is_dir(new_path):
            paths += _url_paths(url + new_path, base_len=base_len)
        else:
            paths.append(url[base_len:] + _replace_spaces(new_path))
    return paths


def find_files(
    uri,
    parent_name,
    outfile,
    gff_pattern=None,
    faa_pattern=None,
    fna_pattern=None,
    exclude=None,
    preference=None,
    name=None,
    name_from_part=None,
    name_split_on=None,
    name_format=None,
    rank=None,
    parent_rank=None,
    write_parent=True,
    parent_only=False,
    download=False,
    nucleic=False,
):
    """Search a URI for FASTA and GFFs to populate an input file."""
    #
    # Process argument list in order: parameters, site defaults, defaults
    #
    if outfile == ():
        outfile = None
    name = list(name)
    if parent_rank is None:
        parent_rank = "species"
    if rank is None:
        rank = "strain"
    if uri.startswith("site://"):  # get defaults from site dict
        splitsite = uri[7:].split("/")
        sitename = splitsite[0]
        if sitename not in SITES:
            logger.error(f'unknown site name "{sitename}"')
            logger.error(f"known sites are {list(SITES.keys())}")
            sys.exit(1)
        sitedict = SITES[sitename]
        uri = sitedict["url"] + "/".join(splitsite[1:])
        if not uri.endswith("/"):
            uri += "/"
        default_gff_pattern = sitedict["gff_pattern"]
        default_faa_pattern = sitedict["faa_pattern"]
        default_fna_pattern = sitedict["fna_pattern"]
        default_name_from_part = sitedict["name_from_part"]
        default_name_split_on = sitedict["name_split_on"]
        default_name_format = sitedict["name_format"]
    else:
        default_gff_pattern = ("*.gf*",)
        default_faa_pattern = ("*.fa*",)
        default_fna_pattern = ("*.fn*",)
        default_name_from_part = 1
        default_name_split_on = "."
        default_name_format = "{0}"
    if gff_pattern == ():
        gff_pattern = default_gff_pattern
    if faa_pattern == ():
        faa_pattern = default_faa_pattern
    if fna_pattern == ():
        fna_pattern = default_fna_pattern
    if name_from_part is None:
        name_from_part = default_name_from_part
    if name_split_on is None:
        name_split_on = default_name_split_on
    if name_format is None:
        name_format = default_name_format
    # TODO-excludes from existing TOML
    if "://" not in uri:
        search_path = Path(uri)
        if not search_path.is_dir():
            logger.error(
                f'path "{search_path}" does not exist or is not a directory'
            )
            sys.exit(1)
        paths = [str(path)[len(uri) :] for path in Path(uri).rglob("*")]
    elif uri.startswith("http"):
        paths = _url_paths(uri)
    else:
        logger.error(f"Badly-formed uri {uri}")
        sys.exit(1)
    if not parent_only:
        last_found = None
        gff_list = []
        faa_list = []
        fna_list = []
        name_list = []
        pref_list = []
        n_warnings = 0
        for path in paths:
            if any([fnmatch(path, exc) for exc in exclude]):
                continue
            if any([fnmatch(path, pat) for pat in gff_pattern]):
                if last_found == "gff":
                    n_warnings += 1
                    logger.warning(
                        f"GFF files adjacent in list: {gff_list[-1]} {path}, adjust gff_pattern"
                    )
                else:
                    last_found = "gff"
                gff_list.append(path)
                if name != []:
                    name_list.append(name.pop())
                else:
                    name_list.append(
                        _path_to_name(
                            path, name_from_part, name_split_on, name_format
                        )
                    )
                pref = None
                if preference != ():
                    for i, pref_pat in enumerate(preference):
                        if fnmatch(path, pref_pat):
                            pref = i
                            break
                pref_list.append(pref)
            if any([fnmatch(path, pat) for pat in faa_pattern]):
                if last_found == "faa":
                    n_warnings += 1
                    logger.warning(
                        f"faa files adjacent in list: {faa_list[-1]} {path}, adjust faa_pattern"
                    )
                else:
                    last_found = "faa"
                faa_list.append(path)
            if nucleic:
                if any([fnmatch(path, pat) for pat in fna_pattern]):
                    if last_found == "fna":
                        n_warnings += 1
                        logger.warning(
                            f"fna files adjacent in list: {fna_list[-1]} {path}, adjust fna_pattern"
                        )
                    else:
                        last_found = "fna"
                    fna_list.append(path)
        if nucleic:
            n_sets = min(len(faa_list), len(gff_list), len(fna_list))
            n_longest = max(len(faa_list), len(gff_list), len(fna_list))
        else:
            n_sets = min(len(faa_list), len(gff_list))
            n_longest = max(len(faa_list), len(gff_list), len(fna_list))
        bad_pattern = n_sets != n_longest
        if outfile is None or n_sets == 0 or bad_pattern:
            logger.info(f"uri: {uri}")
            logger.info(f"gff patterns: {gff_pattern}")
            logger.info(f"faa patterns: {faa_pattern}")
            if nucleic:
                logger.info(f"fna patterns: {fna_pattern}")
            logger.info(f"exclude patterns: {exclude}")
            if n_sets == 0:
                logger.error("No matching sets found")
            else:
                logger.info(f"{n_sets} matching sets found")
            for i in range(n_sets):
                print(f"{i}:")
                print(f"  name: {name_list[i]}")
                print(f"   gff: {gff_list[i]}")
                print(f"   faa: {faa_list[i]}")
                if nucleic:
                    print(f"   fna: {fna_list[i]}")
            if bad_pattern:
                logger.error("Unmatched files:")
                for i in range(n_sets, n_longest):
                    if len(gff_list) > i:
                        gff_path = gff_list[i]
                        name = name_list[i]
                    else:
                        gff_path = "None"
                        name = None
                    logger.error(f"{i}:")
                    logger.error(f"  name: {name}")
                    logger.error(f"   gff: {gff_path}")
                    if len(faa_list) > i:
                        faa_path = faa_list[i]
                    else:
                        faa_path = "None"
                    logger.error(f"   faa: {faa_path}")
                    if nucleic:
                        if len(fna_list) > i:
                            fna_path = fna_list[i]
                        else:
                            fna_path = "None"
                        logger.error(f"   fna: {fna_path}")

                sys.exit(1)
            if n_warnings > 0:
                logger.warning(
                    f"{n_warnings} warnings produced, check results carefully"
                )
            logger.info(f"{n_sets} sets of files found.")
            sys.exit(0)
    #
    # Generate top-level info
    #
    with Path(outfile[0]).open("a+") as outfh:
        if write_parent:
            print(f"[{parent_name}]", file=outfh)
            print(f'rank = "{parent_rank}"\n', file=outfh)
        if parent_only:
            sys.exit(0)
        for i in range(n_sets):
            print(f"[{parent_name}.{name_list[i]}]", file=outfh)
            print(f'rank = "{rank}"', file=outfh)
            print(f'uri = "{uri}"', file=outfh)
            print(f'gff = "{gff_list[i]}"', file=outfh)
            if pref_list[i] != None:
                print(f"preference = {pref_list[i]}", file=outfh)
            if nucleic:
                print(f'fna = "{fna_list[i]}"\n', file=outfh)
            print(f'fasta = "{faa_list[i]}"\n', file=outfh)
